import time
from functools import partial  # needed for the button to sync setpoint with currpoint
import numpy as np

from pyqtgraph.widgets.SpinBox import SpinBox
from qtpy import QtGui, QtWidgets
from qtpy.QtCore import QObject, Slot, QThread, Signal

from simple_pid import PID

from pymodaq.utils.parameter import utils as putils
from pymodaq.utils.parameter import Parameter, ParameterTree
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.daq_utils import ThreadCommand, find_dict_in_list_from_key_val
from pymodaq.utils.managers.modules_manager import ModulesManager
from pymodaq.utils.plotting.data_viewers.viewer0D import Viewer0D
from pymodaq.utils.gui_utils.widgets import QLED
from pymodaq.extensions.pid.utils import DataToActuatorPID, get_models
from pymodaq.utils.gui_utils.dock import DockArea, Dock
from pymodaq.utils.gui_utils.custom_app import CustomApp
from pymodaq.utils.gui_utils.widgets.label import LabelWithFont
from pymodaq.utils.data import DataToExport, DataCalculated, DataActuator, DataRaw
from pymodaq.utils.config import Config

config = Config()
logger = set_logger(get_module_name(__file__))


def convert_output_limits(lim_min=-10., min_status=False, lim_max=10., max_status=False):
    output = [None, None]
    if min_status:
        output[0] = lim_min
    if max_status:
        output[1] = lim_max
    return output


class DAQ_PID(CustomApp):
    """
    """
    command_pid = Signal(ThreadCommand)
    curr_points_signal = Signal(dict)
    setpoints_signal = Signal(dict)
    emit_curr_points_sig = Signal()

    models = get_models()

    params = [
        {'title': 'Models', 'name': 'models', 'type': 'group', 'expanded': True, 'visible': True, 'children': [
            {'title': 'Models class:', 'name': 'model_class', 'type': 'list',
             'limits': [d['name'] for d in models]},
            {'title': 'Model params:', 'name': 'model_params', 'type': 'group', 'children': []},
        ]},
        {'title': 'Move settings:', 'name': 'move_settings', 'expanded': True, 'type': 'group', 'visible': False,
         'children': [
             {'title': 'Units:', 'name': 'units', 'type': 'str', 'value': ''}]},
        # here only to be compatible with DAQ_Scan, the model could update it

        {'title': 'Main Settings:', 'name': 'main_settings', 'expanded': True, 'type': 'group', 'children': [
            {'title': 'Acquisition Timeout (ms):', 'name': 'timeout', 'type': 'int', 'value': 10000},
            {'title': 'epsilon', 'name': 'epsilon', 'type': 'float', 'value': 0.01,
             'tooltip': 'Precision at which move is considered as done'},
            {'title': 'PID controls:', 'name': 'pid_controls', 'type': 'group', 'children': [
                {'title': 'Sample time (ms):', 'name': 'sample_time', 'type': 'int', 'value': 10},
                {'title': 'Refresh plot time (ms):', 'name': 'refresh_plot_time', 'type': 'int', 'value': 200},
                {'title': 'Output limits:', 'name': 'output_limits', 'expanded': True, 'type': 'group', 'children': [
                    {'title': 'Output limit (min):', 'name': 'output_limit_min_enabled', 'type': 'bool',
                     'value': False},
                    {'title': 'Output limit (min):', 'name': 'output_limit_min', 'type': 'float', 'value': 0},
                    {'title': 'Output limit (max):', 'name': 'output_limit_max_enabled', 'type': 'bool',
                     'value': False},
                    {'title': 'Output limit (max:', 'name': 'output_limit_max', 'type': 'float', 'value': 100},
                ]},
                {'title': 'Auto mode:', 'name': 'auto_mode', 'type': 'bool', 'value': False, 'readonly': True},
                {'title': 'Prop. on measurement:', 'name': 'proportional_on_measurement', 'type': 'bool',
                 'value': False},
                {'title': 'PID constants:', 'name': 'pid_constants', 'type': 'group', 'children': [
                    {'title': 'Kp:', 'name': 'kp', 'type': 'float', 'value': 0.1, 'min': 0},
                    {'title': 'Ki:', 'name': 'ki', 'type': 'float', 'value': 0.01, 'min': 0},
                    {'title': 'Kd:', 'name': 'kd', 'type': 'float', 'value': 0.001, 'min': 0},
                ]},

            ]},

        ]},
    ]

    def __init__(self, dockarea, dashboard):
        super().__init__(dockarea, dashboard)

        self.settings = Parameter.create(title='PID settings', name='pid_settings', type='group', children=self.params)
        self.title = 'PyMoDAQ PID'

        self.initialized_state = False
        self.model_class = None
        self._curr_points = dict([])
        self._setpoints = dict([])

        self.dock_area = dockarea
        self.check_moving = False
        self.setup_ui()

        self.enable_controls_pid(False)

        self.enable_controls_pid_run(False)

        self.emit_curr_points_sig.connect(self.emit_curr_points)

    def ini_PID(self):

        if self.is_action_checked('ini_pid'):
            output_limits = [None, None]
            if self.settings['main_settings', 'pid_controls', 'output_limits', 'output_limit_min_enabled']:
                output_limits[0] = self.settings['main_settings', 'pid_controls', 'output_limits', 'output_limit_min']
            if self.settings['main_settings', 'pid_controls', 'output_limits', 'output_limit_max_enabled']:
                output_limits[1] = self.settings['main_settings', 'pid_controls', 'output_limits', 'output_limit_max']

            self.PIDThread = QThread()
            pid_runner = PIDRunner(self.model_class, self.modules_manager, setpoints=self.setpoints,
                                   params=dict(Kp=self.settings['main_settings', 'pid_controls', 'pid_constants',
                                                                      'kp'],
                                               Ki=self.settings['main_settings', 'pid_controls', 'pid_constants',
                                                                      'ki'],
                                               Kd=self.settings['main_settings', 'pid_controls', 'pid_constants',
                                                                      'kd'],
                                               sample_time=self.settings['main_settings', 'pid_controls',
                                                                               'sample_time'] / 1000,
                                               output_limits=output_limits,
                                               auto_mode=False),
                                   )

            self.PIDThread.pid_runner = pid_runner
            pid_runner.pid_output_signal.connect(self.process_output)
            pid_runner.status_sig.connect(self.thread_status)
            self.command_pid.connect(pid_runner.queue_command)

            pid_runner.moveToThread(self.PIDThread)

            self.PIDThread.start()
            self.get_action('pid_led').set_as_true()
            self.enable_controls_pid_run(True)

        else:
            if hasattr(self, 'PIDThread'):
                if self.PIDThread.isRunning():
                    try:
                        self.PIDThread.quit()
                    except Exception:
                        pass
            self.get_action('pid_led').set_as_false()
            self.enable_controls_pid_run(False)

        self.initialized_state = True

    def process_output(self, data: DataToExport):
        inputs: DataRaw = data.get_data_from_name('inputs')
        outputs: DataRaw = data.get_data_from_name('outputs')
        self.curr_points = [float(d) for d in inputs]
        self.output_viewer.show_data(outputs)
        self.input_viewer.show_data(inputs)

    def enable_controls_pid(self, enable=False):
        self.set_action_enabled('ini_pid', enable)
        # self.setpoint_sb.setOpts(enabled=enable)

    def enable_controls_pid_run(self, enable=False):
        self.set_action_enabled('run', enable)
        self.set_action_enabled('pause', enable)

    def setup_menu(self):
        '''
        to be subclassed
        create menu for actions contained into the self.actions_manager, for instance:

        For instance:

        file_menu = self.menubar.addMenu('File')
        self.actions_manager.affect_to('load', file_menu)
        self.actions_manager.affect_to('save', file_menu)

        file_menu.addSeparator()
        self.actions_manager.affect_to('quit', file_menu)
        '''
        pass

    def value_changed(self, param):
        ''' to be subclassed for actions to perform when one of the param's value in self.settings is changed

        For instance:
        if param.name() == 'do_something':
            if param.value():
                print('Do something')
                self.settings.child('main_settings', 'something_done').setValue(False)

        Parameters
        ----------
        param: (Parameter) the parameter whose value just changed
        '''
        if param.name() == 'model_class':
            self.get_set_model_params(param.value())

        elif param.name() == 'refresh_plot_time' or param.name() == 'timeout':
            self.command_pid.emit(ThreadCommand('update_timer', [param.name(), param.value()]))

        elif param.name() == 'sample_time':
            self.command_pid.emit(ThreadCommand('update_options', dict(sample_time=param.value())))

        elif param.name() in putils.iter_children(
                self.settings.child('main_settings', 'pid_controls', 'output_limits'), []):

            output_limits = convert_output_limits(
                self.settings['main_settings', 'pid_controls', 'output_limits',
                              'output_limit_min'],
                self.settings['main_settings', 'pid_controls', 'output_limits',
                              'output_limit_min_enabled'],
                self.settings['main_settings', 'pid_controls', 'output_limits',
                              'output_limit_max'],
                self.settings['main_settings', 'pid_controls', 'output_limits',
                              'output_limit_max_enabled'])

            self.command_pid.emit(ThreadCommand('update_options', dict(output_limits=output_limits)))

        elif param.name() in putils.iter_children(
                self.settings.child('main_settings', 'pid_controls', 'pid_constants'), []):
            Kp = self.settings['main_settings', 'pid_controls', 'pid_constants', 'kp']
            Ki = self.settings['main_settings', 'pid_controls', 'pid_constants', 'ki']
            Kd = self.settings['main_settings', 'pid_controls', 'pid_constants', 'kd']
            self.command_pid.emit(ThreadCommand('update_options', dict(tunings=(Kp, Ki, Kd))))

        elif param.name() in putils.iter_children(self.settings.child('models', 'model_params'), []):
            if self.model_class is not None:
                self.model_class.update_settings(param)

        elif param.name() == 'detector_modules':
            self.model_class.update_detector_names()

    def connect_things(self):
        logger.debug('connecting actions and other')
        self.connect_action('quit', self.quit_fun, )
        self.connect_action('ini_model', self.ini_model)
        self.connect_action('ini_pid', self.ini_PID)
        self.connect_action('run', self.run_PID)
        self.connect_action('pause', self.pause_PID)
        logger.debug('connecting done')

    def setup_actions(self):
        logger.debug('setting actions')
        self.add_action('quit', 'Quit', 'close2', "Quit program")
        self.add_widget('model_label', QtWidgets.QLabel, 'Init Model:')
        self.add_action('ini_model', 'Init Model', 'ini', tip='Initialize the selected model: algo/data conversion')
        self.add_widget('model_led', QLED, toolbar=self.toolbar)
        self.add_widget('model_label', QtWidgets.QLabel, 'Init PID Runner:')
        self.add_action('ini_pid', 'Init the PID loop', 'ini', tip='Init the PID thread', checkable=True)
        self.add_widget('pid_led', QLED, toolbar=self.toolbar)
        self.add_action('run', 'Run The PID loop', 'run2', tip='run or stop the pid loop', checkable=True)
        self.add_action('pause', 'Pause the PID loop', 'pause', tip='Pause the PID loop', checkable=True)
        self.set_action_checked('pause', True)
        logger.debug('actions set')

    def setup_docks(self):
        logger.debug('settings the extension docks')
        self.dock_pid = Dock('PID controller', self.dock_area)
        self.dock_area.addDock(self.dock_pid)

        widget = QtWidgets.QWidget()
        widget_toolbar = QtWidgets.QWidget()
        verlayout = QtWidgets.QVBoxLayout()
        widget.setLayout(verlayout)
        self.toolbar_layout = QtWidgets.QGridLayout()
        widget_toolbar.setLayout(self.toolbar_layout)

        logger.debug('settings the extension docks done')

        labmaj = QtWidgets.QLabel('Sync Value:')
        self.toolbar_layout.addWidget(labmaj, 5, 0, 1, 2)

        verlayout.addWidget(widget_toolbar)
        verlayout.addWidget(self.settings_tree)

        self.dock_output = Dock('PID output')
        widget_output = QtWidgets.QWidget()
        self.output_viewer = Viewer0D(widget_output)
        self.dock_output.addWidget(widget_output)
        self.dock_area.addDock(self.dock_output, 'right', self.dock_pid)

        self.dock_input = Dock('PID input')
        widget_input = QtWidgets.QWidget()
        self.input_viewer = Viewer0D(widget_input)
        self.dock_input.addWidget(widget_input)
        self.dock_area.addDock(self.dock_input, 'bottom', self.dock_output)

        if len(self.models) != 0:
            self.get_set_model_params(self.models[0]['name'])

        self.dock_pid.addWidget(widget)

    def get_set_model_params(self, model_name):
        self.settings.child('models', 'model_params').clearChildren()
        models = get_models()
        if len(models) > 0:
            model_class = find_dict_in_list_from_key_val(models, 'name', model_name)['class']
            params = getattr(model_class, 'params')
            self.settings.child('models', 'model_params').addChildren(params)

    def run_PID(self):
        if self.is_action_checked('run'):
            self.get_action('run').set_icon('stop')
            self.command_pid.emit(ThreadCommand('start_PID', []))
            QtWidgets.QApplication.processEvents()

            QtWidgets.QApplication.processEvents()

            self.command_pid.emit(ThreadCommand('run_PID', [self.model_class.curr_output]))
        else:
            self.get_action('run').set_icon('run2')
            self.command_pid.emit(ThreadCommand('stop_PID'))

            QtWidgets.QApplication.processEvents()

    def pause_PID(self):
        for setp in self.setpoints_sb:
            setp.setEnabled(not self.is_action_checked('pause'))
        self.command_pid.emit(ThreadCommand('pause_PID', [self.is_action_checked('pause')]))

    def stop_moves(self, overshoot):
        """
            Foreach module of the move module object list, stop motion.

            See Also
            --------
            stop_scan,  DAQ_Move_main.daq_move.stop_Motion
        """
        self.overshoot = overshoot
        for mod in self.modules_manager.actuators:
            mod.stop_Motion()

    def set_model(self):
        model_name = self.settings['models', 'model_class']
        self.model_class = find_dict_in_list_from_key_val(self.models, 'name', model_name)['class'](self)
        self.set_setpoints_buttons()
        self.model_class.ini_model()
        self.settings.child('main_settings', 'epsilon').setValue(self.model_class.epsilon)

    def ini_model(self):
        try:
            if self.model_class is None:
                self.set_model()

            self.modules_manager.selected_actuators_name = self.model_class.actuators_name
            self.modules_manager.selected_detectors_name = self.model_class.detectors_name

            self.enable_controls_pid(True)
            self.get_action('model_led').set_as_true()
            self.set_action_enabled('ini_model', False)

        except Exception as e:
            logger.exception(str(e))

    @property
    def setpoints(self):
        return [sp.value() for sp in self.setpoints_sb]

    @setpoints.setter
    def setpoints(self, values):
        for ind, sp in enumerate(self.setpoints_sb):
            sp.setValue(values[ind])

    def setpoints_external(self, values_dict):
        for key in values_dict:
            index = self.model_class.setpoints_names.index(key)
            self.setpoints_sb[index].setValue(values_dict[key])

    @property
    def curr_points(self):
        return [sp.value() for sp in self.currpoints_sb]

    @curr_points.setter
    def curr_points(self, values):
        for ind, sp in enumerate(self.currpoints_sb):
            sp.setValue(values[ind])

    def emit_curr_points(self):
        if self.model_class is not None:
            self.curr_points_signal.emit(dict(zip(self.model_class.setpoints_names, self.curr_points)))

    def set_setpoints_buttons(self):
        self.setpoints_sb = []
        self.currpoints_sb = []
        self.syncvalue_pb = []
        for ind_set in range(self.model_class.Nsetpoints):

            label = LabelWithFont(self.model_class.setpoints_names[ind_set],
                                  font_name="Tahoma", font_size=14, isbold=True, isitalic=True)

            self.setpoints_sb.append(SpinBox())
            self.setpoints_sb[-1].setMinimumHeight(40)
            font = self.setpoints_sb[-1].font()
            font.setPointSizeF(20)
            self.setpoints_sb[-1].setFont(font)
            self.setpoints_sb[-1].setDecimals(6)
            self.toolbar_layout.addWidget(label, 2, 2 + ind_set, 1, 1)
            self.toolbar_layout.addWidget(self.setpoints_sb[-1], 3, 2 + ind_set, 1, 1)
            self.setpoints_sb[-1].valueChanged.connect(self.update_runner_setpoints)

            self.currpoints_sb.append(SpinBox())
            self.currpoints_sb[-1].setMinimumHeight(40)
            self.currpoints_sb[-1].setReadOnly(True)
            self.currpoints_sb[-1].setDecimals(6)
            self.currpoints_sb[-1].setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
            font = self.currpoints_sb[-1].font()
            font.setPointSizeF(20)
            self.currpoints_sb[-1].setFont(font)
            self.toolbar_layout.addWidget(self.currpoints_sb[-1], 4, 2 + ind_set, 1, 1)

            self.syncvalue_pb.append(QtWidgets.QPushButton('Synchro {}'.format(ind_set)))
            self.syncvalue_pb[ind_set].clicked.connect(partial(self.currpoint_as_setpoint, ind_set))
            self.toolbar_layout.addWidget(self.syncvalue_pb[-1], 5, 2 + ind_set)
        self.setpoints_signal.connect(self.setpoints_external)

    def currpoint_as_setpoint(self, i=0):
        '''
        Function used by the sync buttons. The button i will attribute the value of the i-th currpoint to the i-th setpoint.
        '''
        self.setpoints_sb[i].setValue(self.curr_points[i])
        self.update_runner_setpoints

    def quit_fun(self):
        """
        """
        try:
            try:
                self.PIDThread.exit()
            except Exception as e:
                print(e)

            areas = self.dock_area.tempAreas[:]
            for area in areas:
                area.win.close()
                QtWidgets.QApplication.processEvents()
                QThread.msleep(1000)
                QtWidgets.QApplication.processEvents()

            self.dock_area.parent().close()

        except Exception as e:
            print(e)

    def update_runner_setpoints(self):
        self.command_pid.emit(ThreadCommand('update_setpoints', self.setpoints))

    @Slot(list)
    def thread_status(self, status):  # general function to get datas/infos from all threads back to the main
        """

        """
        pass


class PIDRunner(QObject):
    status_sig = Signal(list)
    pid_output_signal = Signal(DataToExport)

    def __init__(self, model_class, modules_manager: ModulesManager, setpoints=[], params=dict([])):
        """
        Init the PID instance with params as initial conditions

        Parameters
        ----------
        params: (dict) Kp=1.0, Ki=0.0, Kd=0.0,setpoints=[0], sample_time=0.01, output_limits=(None, None),
                 auto_mode=True,
                 proportional_on_measurement=False)
        """
        super().__init__()
        self.model_class = model_class
        self.modules_manager = modules_manager
        Nsetpoints = model_class.Nsetpoints
        self.current_time = 0
        self.inputs_from_dets = DataToExport('inputs', data=[DataCalculated(self.model_class.setpoints_names[ind],
                                                                            data=[np.array([setpoints[ind]])])
                                                             for ind in range(Nsetpoints)])
        self.outputs = [0. for _ in range(Nsetpoints)]
        self.outputs_to_actuators = DataToActuatorPID('pid',
                                                      mode='rel',
                                                      data=[DataActuator(self.model_class.actuators_name[ind],
                                                                         data=self.outputs[ind])
                                                            for ind in range(Nsetpoints)])

        if 'sample_time' in params:
            self.sample_time = params['sample_time']
        else:
            self.sample_time = 0.010  # in secs

        self.pids = [PID(setpoint=setpoints[0], **params) for ind in range(Nsetpoints)]  # #PID(object):
        for pid in self.pids:
            pid.set_auto_mode(False)
        self.refreshing_ouput_time = 200
        self.running = True
        self.timer = self.startTimer(self.refreshing_ouput_time)

        self.paused = True

    #     self.timeout_timer = QtCore.QTimer()
    #     self.timeout_timer.setInterval(10000)
    #     self.timeout_scan_flag = False
    #     self.timeout_timer.timeout.connect(self.timeout)
    #
    def timerEvent(self, event):
        dte = DataToExport('toplot', data=[self.outputs_to_actuators.merge_as_dwa('Data0D', name='outputs')])
        dte.append(self.inputs_from_dets.merge_as_dwa('Data0D', name='inputs'))
        self.pid_output_signal.emit(dte)

    @Slot(ThreadCommand)
    def queue_command(self, command: ThreadCommand):
        """
        """
        if command.command == "start_PID":
            self.start_PID(*command.attribute)

        elif command.command == "run_PID":
            self.run_PID(*command.attribute)

        elif command.command == "pause_PID":
            self.pause_PID(*command.attribute)

        elif command.command == "stop_PID":
            self.stop_PID()

        elif command.command == 'update_options':
            self.set_option(**command.attribute)

        elif command.command == 'update_setpoints':
            self.update_setpoints(command.attribute)

        elif command.command == 'input':
            self.update_input(*command.attribute)

        elif command.command == 'update_timer':
            if command.attribute[0] == 'refresh_plot_time':
                self.killTimer(self.timer)
                self.refreshing_ouput_time = command.attribute[1]
                self.timer = self.startTimer(self.refreshing_ouput_time)
            elif command.attribute[0] == 'timeout':
                self.timeout_timer.setInterval(command.attribute[1])

    def update_input(self, measurements: DataToExport):
        self.inputs_from_dets = self.model_class.convert_input(measurements)

    def start_PID(self, sync_detectors=True, sync_acts=False):
        """Start the pid controller loop

        Parameters
        ----------
        sync_detectors: (bool) if True will make sure all selected detectors (if any) all got their data before calling
            the model
        sync_acts: (bool) if True will make sure all selected actuators (if any) all reached their target position
         before calling the model
        """
        self.running = True
        try:
            if sync_detectors:
                self.modules_manager.connect_detectors()
            if sync_acts:
                self.modules_manager.connect_actuators()

            self.current_time = time.perf_counter()
            logger.info('PID loop starting')
            while self.running:
                # print('input: {}'.format(self.input))
                # # GRAB DATA FIRST AND WAIT ALL DETECTORS RETURNED

                self.det_done_datas: DataToExport = self.modules_manager.grab_datas()

                self.inputs_from_dets: DataToExport = self.model_class.convert_input(self.det_done_datas)

                # # EXECUTE THE PID
                self.outputs = []
                for ind, pid in enumerate(self.pids):
                    self.outputs.append(pid(float(self.inputs_from_dets[ind][0][0])))

                # # APPLY THE PID OUTPUT TO THE ACTUATORS
                if self.outputs is None:
                    self.outputs = [pid.setpoint for pid in self.pids]

                dt = time.perf_counter() - self.current_time
                self.outputs_to_actuators: DataToActuatorPID = self.model_class.convert_output(self.outputs, dt,
                                                                                               stab=True)

                if not self.paused:
                    self.modules_manager.move_actuators(self.outputs_to_actuators,
                                                        self.outputs_to_actuators.mode,
                                                        polling=False)

                self.current_time = time.perf_counter()
                QtWidgets.QApplication.processEvents()
                QThread.msleep(int(self.sample_time * 1000))

            logger.info('PID loop exiting')
            self.modules_manager.connect_actuators(False)
            self.modules_manager.connect_detectors(False)

        except Exception as e:
            logger.exception(str(e))

    def update_setpoints(self, setpoints):
        for ind, pid in enumerate(self.pids):
            pid.setpoint = setpoints[ind]

    def set_option(self, **option):
        for pid in self.pids:
            for key in option:
                if hasattr(pid, key):
                    if key == 'sample_time':
                        setattr(pid, key, option[key] / 1000)
                    else:
                        setattr(pid, key, option[key])

    def run_PID(self, last_values):
        logger.info('Stabilization started')
        for ind, pid in enumerate(self.pids):
            pid.set_auto_mode(True, last_values[ind])

    def pause_PID(self, pause_state):
        for ind, pid in enumerate(self.pids):
            if pause_state:
                pid.set_auto_mode(False)
                logger.info('Stabilization paused')
            else:
                pid.set_auto_mode(True, self.outputs[ind])
                logger.info('Stabilization restarted from pause')
        self.paused = pause_state

    def stop_PID(self):
        self.running = False
        logger.info('PID loop exiting')


if __name__ == '__main__':
    import sys
    from pathlib import Path
    from pymodaq.utils.daq_utils import get_set_preset_path

    app = QtWidgets.QApplication(sys.argv)
    if config['style']['darkstyle']:
        import qdarkstyle
        app.setStyleSheet(qdarkstyle.load_stylesheet())

    from pymodaq.dashboard import DashBoard

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Dashboard')

    dashboard = DashBoard(area)
    pid = None
    file = Path(get_set_preset_path()).joinpath(f"{config('presets', 'default_preset_for_pid')}.xml")
    if file.exists():
        dashboard.set_preset_mode(file)
        pid = dashboard.load_pid_module()
    else:
        msgBox = QtWidgets.QMessageBox()
        msgBox.setText(f"The default file specified in the configuration file does not exists!\n"
                       f"{file}\n"
                       f"Impossible to load the DAQ_PID Module")
        msgBox.setStandardButtons(msgBox.Ok)
        ret = msgBox.exec()

    sys.exit(app.exec_())








# -*- coding: utf-8 -*-
"""
Created the 15/11/2022

@author: Sebastien Weber
"""
import copy
import datetime
from dateutil import parser
from numbers import Number
import os
from pathlib import Path
from typing import Union, Iterable


import numpy as np
from qtpy.QtCore import QObject, Signal
from qtpy import QtWidgets

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.parameter import Parameter, ParameterTree
from pymodaq.utils.parameter import utils as putils
from pymodaq.utils.managers.parameter_manager import ParameterManager
from pymodaq.utils.gui_utils.file_io import select_file

from pymodaq.utils.gui_utils.utils import dashboard_submodules_params
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.config import Config
from pymodaq.utils.data import DataDim, DataToExport, Axis, DataWithAxes
from pymodaq.utils.enums import BaseEnum, enum_checker
from pymodaq.utils.scanner.utils import ScanType
from pymodaq.utils.messenger import deprecation_msg


from .backends import (H5Backend, backends_available, SaveType, InvalidSave, InvalidExport, InvalidDataType,
                       InvalidGroupType, InvalidGroupDataType, Node, GroupType, InvalidDataDimension, InvalidScanType,
                       GROUP, VLARRAY)
from . import browsing


config = Config()
logger = set_logger(get_module_name(__file__))


class FileType(BaseEnum):
    detector = 0
    actuator = 1
    axis = 2
    scan = 3
    

class DataType(BaseEnum):
    data = 'Data'
    axis = 'Axis'
    live_scan = 'Live'
    external_h5 = 'ExtData'
    strings = 'Strings'
    bkg = 'Bkg'
    data_enlargeable = 'EnlData'


class H5SaverLowLevel(H5Backend):
    """Object containing basic methods in order to structure and interact with a h5file compatible with the h5browser

    See Also
    --------
    H5Browser

    Attributes
    ----------
    h5_file: pytables hdf5 file
        object used to save all datas and metadas
    h5_file_path: str or Path
        The file path
    """

    def __init__(self, save_type: SaveType = 'scan', backend='tables'):
        H5Backend.__init__(self, backend)

        self.save_type = enum_checker(SaveType, save_type)

        self.h5_file_path = None
        self.h5_file_name = None
        self.file_loaded = False

        self._current_group = None
        self._raw_group: Union[GROUP, str] = '/RawData'
        self._logger_array = None

    @property
    def raw_group(self):
        return self._raw_group

    @property
    def h5_file(self):
        return self._h5file

    def init_file(self, file_name: Path, raw_group_name='RawData', new_file=False, metadata: dict = None):
        """Initializes a new h5 file.

        Parameters
        ----------
        file_name: Path
            a complete Path pointing to a h5 file
        raw_group_name: str
            Base node name
        new_file: bool
            If True create a new file, otherwise append to a potential existing one

        Returns
        -------
        bool
            True if new file has been created, False otherwise
        """
        datetime_now = datetime.datetime.now()

        if file_name is not None and isinstance(file_name, Path):
            self.h5_file_name = file_name.stem + ".h5"
            self.h5_file_path = file_name.parent
            if not self.h5_file_path.joinpath(self.h5_file_name).is_file():
                new_file = True

        else:
            self.h5_file_name = select_file(save=True, ext='h5')
            self.h5_file_path = self.h5_file_name.parent
            new_file = True

        self.close_file()
        self.open_file(self.h5_file_path.joinpath(self.h5_file_name), 'w' if new_file else 'a', title='PyMoDAQ file')

        self._raw_group = self.get_set_group(self.root(), raw_group_name, title='Data from PyMoDAQ modules')
        self.get_set_logger(self._raw_group)

        if new_file:
            self._raw_group.attrs['type'] = self.save_type.name  # first possibility to set a node attribute
            self.root().set_attr('file', self.h5_file_name)  # second possibility

            self.set_attr(self.root(), 'date', datetime_now.date().isoformat())
            self.set_attr(self.root(), 'time', datetime_now.time().isoformat())

            if metadata is not None:
                for metadata_key in metadata:
                    self._raw_group.attrs[metadata_key] = metadata[metadata_key]

    def save_file(self, filename=None):
        if filename is None:
            filename = select_file(None, save=True, ext='h5')
        if filename != '':
            super().save_file_as(filename)

    def get_set_logger(self, where: Node = None) -> VLARRAY:
        """ Retrieve or create (if absent) a logger enlargeable array to store logs
        Get attributed to the class attribute ``logger_array``
        Parameters
        ----------
        where: node
               location within the tree where to save or retrieve the array

        Returns
        -------
        vlarray
            enlargeable array accepting strings as elements
        """
        if where is None:
            where = self.raw_group
        if isinstance(where, Node):
            where = where.node
        logger = 'Logger'
        if logger not in list(self.get_children(where)):
            # check if logger node exist
            self._logger_array = self.add_string_array(where, logger)
            self._logger_array.attrs['type'] = 'log'
        else:
            self._logger_array = self.get_node(where, name=logger)
        return self._logger_array

    def add_log(self, msg):
        self._logger_array.append(msg)

    def add_string_array(self, where, name, title='', metadata=dict([])):
        array = self.create_vlarray(where, name, dtype='string', title=title)
        array.attrs['shape'] = (0,)
        array.attrs['data_type'] = 'strings'

        for metadat in metadata:
            array.attrs[metadat] = metadata[metadat]
        return array
    
    def add_array(self, where: Union[GROUP, str], name: str, data_type: DataType, array_to_save: np.ndarray = None,
                  data_shape: tuple = None, array_type: np.dtype = None, data_dimension: DataDim = None,
                  scan_shape: tuple = tuple([]), add_scan_dim=False, enlargeable: bool = False,
                  title: str = '', metadata=dict([]), ):

        """save data arrays on the hdf5 file together with metadata
        Parameters
        ----------
        where: GROUP
            node where to save the array
        name: str
            name of the array in the hdf5 file
        data_type: DataType
            mandatory so that the h5Browser can interpret correctly the array
        data_shape: Iterable
            the shape of the array to save, mandatory if array_to_save is None
        data_dimension: DataDim
         The data's dimension
        scan_shape: Iterable
            the shape of the scan dimensions
        title: str
            the title attribute of the array node
        array_to_save: ndarray or None
            data to be saved in the array. If None, array_type and data_shape should be specified in order to init
            correctly the memory
        array_type: np.dtype or numpy types
            eg np.float, np.int32 ...
        enlargeable: bool
            if False, data are saved as a CARRAY, otherwise as a EARRAY (for ragged data, see add_string_array)
        metadata: dict
            dictionnary whose keys will be saved as the array attributes
        add_scan_dim: if True, the scan axes dimension (scan_shape iterable) is prepended to the array shape on the hdf5
                      In that case, the array is usually initialized as zero and further populated

        Returns
        -------
        array (CARRAY or EARRAY)

        See Also
        --------
        add_data, add_string_array
        """
        if array_type is None:
            if array_to_save is None:
                array_type = config('data_saving', 'data_type', 'dynamic')
            else:
                array_type = array_to_save.dtype

        data_type = enum_checker(DataType, data_type)
        data_dimension = enum_checker(DataDim, data_dimension)

        if enlargeable:
            # if data_shape == (1,):
            #     data_shape = None
            array = self.create_earray(where, utils.capitalize(name), dtype=np.dtype(array_type),
                                       data_shape=data_shape, title=title)
        else:
            if add_scan_dim:  # means it is an array initialization to zero
                shape = list(scan_shape[:])
                if not(len(data_shape) == 1 and data_shape[0] == 1):  # means data are not ndarrays of scalars
                    shape.extend(data_shape)
                if array_to_save is None:
                    array_to_save = np.zeros(shape, dtype=np.dtype(array_type))

            array = self.create_carray(where, utils.capitalize(name), obj=array_to_save, title=title)
        self.set_attr(array, 'data_type', data_type.name)
        self.set_attr(array, 'data_dimension', data_dimension.name)

        for metadat in metadata:
            self.set_attr(array, metadat, metadata[metadat])
        return array

    def get_set_group(self, where, name, title=''):
        """Get the group located at where if it exists otherwise creates it

        This also set the _current_group property
        """
        self._current_group = super().get_set_group(where, name, title)
        return self._current_group

    def get_groups(self, where: Union[str, GROUP], group_type: GroupType):
        """Get all groups hanging from a Group and of a certain type"""
        groups = []
        for node_name in list(self.get_children(where)):
            group = self.get_node(where, node_name)
            if 'type' in group.attrs and group.attrs['type'] == group_type.name:
                groups.append(group)
        return groups

    def get_last_group(self, where: GROUP, group_type: GroupType):
        groups = self.get_groups(where, group_type)
        if len(groups) != 0:
            return groups[-1]
        else:
            return None

    def get_node_from_attribute_match(self, where, attr_name, attr_value):
        """Get a Node starting from a given node (Group) matching a pair of node attribute name and value"""
        for node in self.walk_nodes(where):
            if attr_name in node.attrs and node.attrs[attr_name] == attr_value:
                return node

    def get_node_from_title(self, where, title: str):
        """Get a Node starting from a given node (Group) matching the given title"""
        return self.get_node_from_attribute_match(where, 'TITLE', title)

    def add_data_group(self, where, data_dim: DataDim, title='', settings_as_xml='', metadata=dict([])):
        """Creates a group node at given location in the tree

        Parameters
        ----------
        where: group node
               where to create data group
        group_data_type: DataDim
        title: str, optional
               a title for this node, will be saved as metadata
        settings_as_xml: str, optional
                         XML string created from a Parameter object to be saved as metadata
        metadata: dict, optional
                  will be saved as a new metadata attribute with name: key and value: dict value

        Returns
        -------
        group: group node

        See Also
        --------
        :py:meth:`add_group`
        """
        data_dim = enum_checker(DataDim, data_dim)
        metadata.update(settings=settings_as_xml)
        group = self.add_group(data_dim.name, 'data_dim', where, title, metadata)
        return group

    def add_incremental_group(self, group_type, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a node in the h5 file tree of the group type with an increment in the given name
        Parameters
        ----------
        group_type: str or GroupType enum
            one of the possible values of **group_types**
        where: str or node
            parent node where to create the new group
        title: str
            node title
        settings_as_xml: str
            XML string containing Parameter representation
        metadata: dict
            extra metadata to be saved with this new group node

        Returns
        -------
        node: newly created group node
        """
        group_type = enum_checker(GroupType, group_type)

        nodes = [name for name in self.get_children(self.get_node(where))]
        nodes_tmp = []
        for node in nodes:
            if utils.capitalize(group_type.name) in node:
                nodes_tmp.append(node)
        nodes_tmp.sort()
        if len(nodes_tmp) == 0:
            ind_group = -1
        else:
            ind_group = int(nodes_tmp[-1][-3:])
        group = self.get_set_group(where, f'{utils.capitalize(group_type.name)}{ind_group + 1:03d}', title)
        self.set_attr(group, 'settings', settings_as_xml)
        if group_type.name.lower() != 'ch':
            self.set_attr(group, 'type', group_type.name.lower())
        else:
            self.set_attr(group, 'type', '')
        for metadat in metadata:
            self.set_attr(group, metadat, metadata[metadat])
        return group

    def add_act_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type detector
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('actuator', where, title, settings_as_xml, metadata)
        return group

    def add_det_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type detector
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('detector', where, title, settings_as_xml, metadata)
        return group

    def add_scan_group(self, where='/RawData', title='', settings_as_xml='', metadata=dict([])):
        """Add a new group of type scan

        At creation adds the attributes description and scan_done to be used elsewhere

        See Also
        -------
        add_incremental_group
        """
        metadata.update(dict(description='', scan_done=False))
        group = self.add_incremental_group(GroupType['scan'], where, title, settings_as_xml, metadata)
        return group

    def add_ch_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type channel
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('ch', where, title, settings_as_xml, metadata)
        return group


    def add_move_group(self, where, title='', settings_as_xml='', metadata=dict([])):
        """
        Add a new group of type actuator
        See Also
        -------
        add_incremental_group
        """
        group = self.add_incremental_group('actuator', where, title, settings_as_xml, metadata)
        return group

    def show_file_content(self):
        win = QtWidgets.QMainWindow()
        if not self.isopen():
            if self.h5_file_path is not None:
                if self.h5_file_path.exists():
                    self.analysis_prog = browsing.H5Browser(win, h5file_path=self.h5_file_path)
                else:
                    logger.warning('The h5 file path has not been defined yet')
            else:
                logger.warning('The h5 file path has not been defined yet')
        else:
            self.flush()
            self.analysis_prog = browsing.H5Browser(win, h5file=self.h5file)
        win.show()


class H5SaverBase(H5SaverLowLevel, ParameterManager):
    """Object containing all methods in order to save datas in a *hdf5 file* with a hierarchy compatible with
    the H5Browser. The saving parameters are contained within a **Parameter** object: self.settings that can be displayed
    on a UI using the widget self.settings_tree. At the creation of a new file, a node
    group named **Raw_datas** and represented by the attribute ``raw_group`` is created and set with a metadata attribute:

    * 'type' given by the **save_type** class parameter

    The root group of the file is then set with a few metadata:

    * 'pymodaq_version' the current pymodaq version, e.g. 1.6.2
    * 'file' the file name
    * 'date' the current date
    * 'time' the current time

    All datas will then be saved under this node in various groups

    See Also
    --------
    H5Browser

    Parameters
    ----------
    h5_file: pytables hdf5 file
             object used to save all datas and metadas
    h5_file_path: str or Path
                  Signal signal represented by a float. Is emitted each time the hardware reached the target
                  position within the epsilon precision (see comon_parameters variable)
    save_type: str
       an element of the enum module attribute SaveType
       * 'scan' is used for DAQScan module and should be used for similar application
       * 'detector' is used for DAQ_Viewer module and should be used for similar application
       * 'custom' should be used for customized applications

    Attributes
    ----------

    settings: Parameter
               Parameter instance (pyqtgraph) containing all settings (could be represented using the settings_tree widget)

    settings_tree: ParameterTree
                   Widget representing as a Tree structure, all the settings defined in the class preamble variable ``params``

    """
    settings_name = 'h5saver_settings'
    params = [
        {'title': 'Save type:', 'name': 'save_type', 'type': 'list', 'limits': SaveType.names(), 'readonly': True},
    ] + dashboard_submodules_params + \
        [{'title': 'Backend:', 'name': 'backend', 'type': 'group', 'children': [
            {'title': 'Backend type:', 'name': 'backend_type', 'type': 'list', 'limits': backends_available,
                'readonly': True},
            {'title': 'HSDS Server:', 'name': 'hsds_options', 'type': 'group', 'visible': False, 'children': [
                {'title': 'Endpoint:', 'name': 'endpoint', 'type': 'str',
                    'value': config('data_saving', 'hsds', 'root_url'), 'readonly': False},
                {'title': 'User:', 'name': 'user', 'type': 'str',
                    'value': config('data_saving', 'hsds', 'username'), 'readonly': False},
                {'title': 'password:', 'name': 'password', 'type': 'str',
                    'value': config('data_saving', 'hsds', 'pwd'), 'readonly': False},
            ]},
        ]},

        {'title': 'custom_name?:', 'name': 'custom_name', 'type': 'bool', 'default': False, 'value': False},
        {'title': 'show file content?', 'name': 'show_file', 'type': 'bool_push', 'default': False,
            'value': False},
        {'title': 'Base path:', 'name': 'base_path', 'type': 'browsepath',
            'value': config('data_saving', 'h5file', 'save_path'), 'filetype': False, 'readonly': True, },
        {'title': 'Base name:', 'name': 'base_name', 'type': 'str', 'value': 'Scan', 'readonly': True},
        {'title': 'Current scan:', 'name': 'current_scan_name', 'type': 'str', 'value': '', 'readonly': True},
        {'title': 'Current path:', 'name': 'current_scan_path', 'type': 'text',
            'value': config('data_saving', 'h5file', 'save_path'), 'readonly': True, 'visible': False},
        {'title': 'h5file:', 'name': 'current_h5_file', 'type': 'text', 'value': '', 'readonly': True},
        {'title': 'New file', 'name': 'new_file', 'type': 'action'},
        {'title': 'Saving dynamic', 'name': 'dynamic', 'type': 'list',
         'limits': config('data_saving', 'data_type', 'dynamics'),
         'value': config('data_saving', 'data_type', 'dynamic')},
        {'title': 'Compression options:', 'name': 'compression_options', 'type': 'group', 'children': [
            {'title': 'Compression library:', 'name': 'h5comp_library', 'type': 'list', 'value': 'zlib',
                'limits': ['zlib', 'gzip']},
            {'title': 'Compression level:', 'name': 'h5comp_level', 'type': 'int',
                'value': config('data_saving', 'h5file', 'compression_level'), 'min': 0, 'max': 9},
        ]},
    ]

    def __init__(self, save_type='scan', backend='tables'):
        """

        Parameters
        ----------
        save_type (str): one of ['scan', 'detector', 'logger', 'custom']
        backend (str): either 'tables' for pytables backend, 'h5py' for h5py backends or 'h5pyd' for HSDS backend

        See Also
        --------
        https://github.com/HDFGroup/hsds
        """
        H5SaverLowLevel.__init__(self, save_type, backend)
        ParameterManager.__init__(self)

        self.current_scan_group = None
        self.current_scan_name = None

        self.settings.child('save_type').setValue(self.save_type.name)

    def show_settings(self, show=True):
        self.settings_tree.setVisible(show)

    def init_file(self, update_h5=False, custom_naming=False, addhoc_file_path=None, metadata=dict([])):
        """Initializes a new h5 file.
        Could set the h5_file attributes as:

        * a file with a name following a template if ``custom_naming`` is ``False`` and ``addhoc_file_path`` is ``None``
        * a file within a name set using a file dialog popup if ``custom_naming`` is ``True``
        * a file with a custom name if ``addhoc_file_path`` is a ``Path`` object or a path string

        Parameters
        ----------
        update_h5: bool
                   create a new h5 file with name specified by other parameters
                   if false try to open an existing file and will append new data to it
        custom_naming: bool
                       if True, a selection file dialog opens to set a new file name
        addhoc_file_path: Path or str
                          supplied name by the user for the new file
        metadata: dict
                    dictionnary with pair of key, value that should be saved as attributes of the root group
        Returns
        -------
        update_h5: bool
                   True if new file has been created, False otherwise
        """
        datetime_now = datetime.datetime.now()
        if addhoc_file_path is None:
            if not os.path.isdir(self.settings['base_path']):
                os.mkdir(self.settings['base_path'])

            # set the filename and path
            base_name = self.settings['base_name']

            if not custom_naming:
                custom_naming = self.settings['custom_name']

            if not custom_naming:
                scan_type = self.settings['save_type'] == 'scan'
                scan_path, current_scan_name, save_path = self.update_file_paths(update_h5)
                self.current_scan_name = current_scan_name
                self.settings.child('current_scan_name').setValue(current_scan_name)
                self.settings.child('current_scan_path').setValue(str(scan_path))

                if not scan_type:
                    self.h5_file_path = save_path.parent  # will remove the dataset part used for DAQ_scan datas
                    self.h5_file_name = base_name + datetime_now.strftime('_%Y%m%d_%H_%M_%S.h5')
                else:
                    self.h5_file_name = save_path.name + ".h5"
                    self.h5_file_path = save_path.parent

            else:
                self.h5_file_name = select_file(start_path=base_name, save=True, ext='h5')
                self.h5_file_path = self.h5_file_name.parent

        else:
            if isinstance(addhoc_file_path, str):
                addhoc_file_path = Path(addhoc_file_path)
            self.h5_file_path = addhoc_file_path.parent
            self.h5_file_name = addhoc_file_path.name

        fullpathname = self.h5_file_path.joinpath(self.h5_file_name)
        self.settings.child('current_h5_file').setValue(str(fullpathname))

        super().init_file(fullpathname, new_file=update_h5, metadata=metadata)

        self.get_set_logger(self.raw_group)

        return update_h5

    def update_file_paths(self, update_h5=False):
        """

        Parameters
        ----------
        update_h5: bool
                   if True, will increment the file name and eventually the current scan index
                   if False, get the current scan index in the h5 file

        Returns
        -------
        scan_path: Path
        current_filename: str
        dataset_path: Path

        """

        try:
            # set the filename and path
            base_path = self.settings['base_path']
            base_name = self.settings['base_name']
            current_scan = self.settings['current_scan_name']
            scan_type = self.settings['save_type'] == 'scan'
            ind_dataset = None
            if current_scan == '' or update_h5:
                next_scan_index = 0
                update_h5 = True  # just started the main program so one should create a new h5
                self.file_loaded = False
            else:
                next_scan_index = self.get_scan_index()
            if self.file_loaded:
                ind_dataset = int(os.path.splitext(self.h5_file_name)[0][-3:])
                try:
                    curr_date = datetime.date.fromisoformat(self.get_attr(self.root(), 'date'))
                except ValueError:
                    curr_date = parser.parse(self.get_attr(self.root(), 'date')).date()
            else:
                curr_date = datetime.date.today()

            scan_path, current_filename, dataset_path = self.set_current_scan_path(base_path, base_name, update_h5,
                                                                                   next_scan_index,
                                                                                   create_dataset_folder=False,
                                                                                   curr_date=curr_date,
                                                                                   ind_dataset=ind_dataset)
            self.settings.child('current_scan_path').setValue(str(dataset_path))

            return scan_path, current_filename, dataset_path

        except Exception as e:
            logger.exception(str(e))

    @classmethod
    def find_part_in_path_and_subpath(cls, base_dir, part='', create=False, increment=True):
        """
        Find path from part time.

        =============== ============ =============================================
        **Parameters**  **Type**      **Description**
        *base_dir*      Path object   The directory to browse
        *part*          string        The date of the directory to find/create
        *create*        boolean       Indicate the creation flag of the directory
        =============== ============ =============================================

        Returns
        -------
        Path object
            found path from part
        """
        found_path = None
        if part in base_dir.parts:  # check if current year is in the given base path
            if base_dir.name == part:
                found_path = base_dir
            else:
                for ind in range(len(base_dir.parts)):
                    tmp_path = base_dir.parents[ind]
                    if tmp_path.name == part:
                        found_path = base_dir.parents[ind]
                        break
        else:  # if not check if year is in the subfolders
            subfolders_year_name = [x.name for x in base_dir.iterdir() if x.is_dir()]
            subfolders_found_path = [x for x in base_dir.iterdir() if x.is_dir()]
            if part not in subfolders_year_name:
                if increment:
                    found_path = base_dir.joinpath(part)
                else:
                    found_path = base_dir
                if create:
                    found_path.mkdir()
            else:
                ind_path = subfolders_year_name.index(part)
                found_path = subfolders_found_path[ind_path]
        return found_path

    @classmethod
    def set_current_scan_path(cls, base_dir, base_name='Scan', update_h5=False, next_scan_index=0,
                              create_scan_folder=False,
                              create_dataset_folder=True, curr_date=None, ind_dataset=None):
        """

        Parameters
        ----------
        base_dir
        base_name
        update_h5
        next_scan_index
        create_scan_folder
        create_dataset_folder

        Returns
        -------

        """
        base_dir = Path(base_dir)
        if curr_date is None:
            curr_date = datetime.date.today()

        year_path = cls.find_part_in_path_and_subpath(base_dir, part=str(curr_date.year),
                                                      create=True)  # create directory of the year if it doen't exist and return it
        day_path = cls.find_part_in_path_and_subpath(year_path, part=curr_date.strftime('%Y%m%d'),
                                                     create=True)  # create directory of the day if it doen't exist and return it
        dataset_base_name = curr_date.strftime('Dataset_%Y%m%d')
        dataset_paths = sorted([path for path in day_path.glob(dataset_base_name + "*"+".h5") if path.is_file()])

        if ind_dataset is None:
            if dataset_paths == []:

                ind_dataset = 0
            else:
                if update_h5:
                    ind_dataset = int(dataset_paths[-1].stem.partition(dataset_base_name + "_")[2]) + 1
                else:
                    ind_dataset = int(dataset_paths[-1].stem.partition(dataset_base_name + "_")[2])

        dataset_path = cls.find_part_in_path_and_subpath(day_path,
                                                         part=dataset_base_name + "_{:03d}".format(ind_dataset),
                                                         create=False, increment=True)
        scan_paths = sorted([path for path in dataset_path.glob(base_name + '*') if path.is_dir()])
        ind_scan = next_scan_index
        return dataset_path, base_name + '{:03d}'.format(ind_scan), dataset_path

    def get_last_scan(self):
        """Gets the last scan node within the h5_file and under the **raw_group**

        Returns
        -------
        scan_group: pytables group or None


        """
        return self.get_last_group(self.raw_group, GroupType['scan'])

    def get_scan_groups(self):
        return self.get_groups(self.raw_group, GroupType['scan'])

    def get_scan_index(self):
        """ return the scan group index in the "scan templating": Scan000, Scan001 as an integer
        """

        last_scan = self.get_last_scan()
        return int(last_scan.name[4:]) if last_scan is not None else 0

    def load_file(self, base_path=None, file_path=None):
        """Opens a file dialog to select a h5file saved on disk to be used

        Parameters
        ----------
        base_path
        file_path

        See Also
        --------
        :py:meth:`init_file`

        """
        if base_path is None:
            base_path = self.settings.child('base_path').value()
            if not os.path.isdir(base_path):
                base_path = None

        if file_path is None:
            file_path = select_file(base_path, save=False, ext='h5')

        if not (file_path is None or file_path == ''):
            if not isinstance(file_path, Path):
                file_path = Path(file_path)

            if 'h5' not in file_path.suffix:
                raise IOError('Invalid file type, should be a h5 file')

            self.init_file(addhoc_file_path=file_path)
            self.file_loaded = True

    def save_file(self, filename=None):
        if filename is None:
            filename = select_file(None, save=True, ext='h5')
        if filename != '':
            super().save_file_as(filename)

    def value_changed(self, param):
        if param.name() == 'show_file':
            if param.value():
                param.setValue(False)
                self.show_file_content()

        elif param.name() == 'base_path':
            try:
                if not os.path.isdir(param.value()):
                    os.mkdir(param.value())
            except Exception as e:
                self.update_status(f"The base path couldn't be set, please check your options: {str(e)}")

        elif param.name() in putils.iter_children(self.settings.child('compression_options'), []):
            compression = self.settings.child('compression_options', 'h5comp_library').value()
            compression_opts = self.settings.child('compression_options', 'h5comp_level').value()
            self.define_compression(compression, compression_opts)

    def update_status(self, status):
        logger.warning(status)


class H5Saver(H5SaverBase, QObject):
    """
    status_sig: Signal
                emits a signal of type Threadcommand in order to senf log information to a main UI
    new_file_sig: Signal
                  emits a boolean signal to let the program know when the user pressed the new file button on the UI
    """

    status_sig = Signal(utils.ThreadCommand)
    new_file_sig = Signal(bool)

    def __init__(self, *args, **kwargs):
        """

        Parameters
        ----------
        args
        kwargs
        """
        QObject.__init__(self)
        H5SaverBase.__init__(self, *args, **kwargs)

        self.settings.child('new_file').sigActivated.connect(lambda: self.emit_new_file(True))

    def close(self):
        self.close_file()

    def emit_new_file(self, status):
        """Emits the new_file_sig

        Parameters
        ----------
        status: bool
                emits True if a new file has been asked by the user pressing the new file button on the UI
        """
        self.new_file_sig.emit(status)




# -*- coding: utf-8 -*-
"""
Created the 05/12/2022

@author: Sebastien Weber
"""
from typing import List, Tuple

import numpy as np

from qtpy import QtCore, QtWidgets
from pymodaq.utils.data import Axis, DataDistribution
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils import math_utils as mutils
from pymodaq.utils import config as configmod
from pymodaq.utils import gui_utils as gutils
from ..scan_factory import ScannerFactory, ScannerBase, ScanParameterManager
from pymodaq.utils.parameter import utils as putils
from pymodaq.utils.parameter.pymodaq_ptypes import TableViewCustom
from pymodaq.utils.plotting.scan_selector import Selector
from pymodaq.utils.plotting.utils.plot_utils import Point, get_sub_segmented_positions


logger = set_logger(get_module_name(__file__))
config = configmod.Config()


class TableModelTabular(gutils.TableModel):
    """Table Model for the Model/View Qt framework dedicated to the Tabular scan mode"""
    def __init__(self, data, axes_name=None, **kwargs):
        if axes_name is None:
            if 'header' in kwargs:  # when saved as XML the header will be saved and restored here
                axes_name = [h for h in kwargs['header']]
                kwargs.pop('header')
            else:
                raise Exception('Invalid header')

        header = [name for name in axes_name]
        editable = [True for name in axes_name]
        super().__init__(data, header, editable=editable, **kwargs)

    def __len__(self):
        return len(self._data)

    def add_data(self, row, data=None):
        if data is not None:
            self.insert_data(row, [float(d) for d in data])
        else:
            self.insert_data(row, [0. for name in self.header])

    def remove_data(self, row):
        self.remove_row(row)

    def load_txt(self):
        fname = gutils.select_file(start_path=None, save=False, ext='*')
        if fname is not None and fname != '':
            while self.rowCount(self.index(-1, -1)) > 0:
                self.remove_row(0)

            data = np.loadtxt(fname)
            if len(data.shape) == 1:
                data = data.reshape((data.size, 1))
            self.set_data_all(data)

    def save_txt(self):
        fname = gutils.select_file(start_path=None, save=True, ext='dat')
        if fname is not None and fname != '':
            np.savetxt(fname, self.get_data_all(), delimiter='\t')

    def __repr__(self):
        return f'{self.__class__.__name__} from module {self.__class__.__module__}'

    def validate_data(self, row, col, value):
        """
        make sure the values and signs of the start, stop and step values are "correct"
        Parameters
        ----------
        row: (int) row within the table that is to be changed
        col: (int) col within the table that is to be changed
        value: (float) new value for the value defined by row and col

        Returns
        -------
        bool: True is the new value is fine (change some other values if needed) otherwise False
        """

        return True


class TableModelTabularReadOnly(TableModelTabular):
    def setData(self, index, value, role):
        return False


@ScannerFactory.register()
class TabularScanner(ScannerBase):
    scan_type = 'Tabular'
    scan_subtype = 'Linear'
    save_settings = False  # not easy to save table content in a toml...
    params = [
        {'title': 'Positions', 'name': 'tabular_table', 'type': 'table_view', 'delegate': gutils.SpinBoxDelegate,
         'menu': True},
              ]
    distribution = DataDistribution['spread']

    def __init__(self, actuators: List[str]):
        self.table_model: TableModelTabular = None
        self.table_view: TableViewCustom = None
        super().__init__(actuators=actuators)
        self.update_model()

    @property
    def actuators(self):
        return self._actuators

    @actuators.setter
    def actuators(self, actuators_name):
        self._actuators = actuators_name
        self.update_model()

    def update_model(self, init_data=None):
        if init_data is None:
            init_data = [[0. for _ in self._actuators]]

        self.table_model = TableModelTabular(init_data, [act.title for act in self._actuators])
        self.table_view = putils.get_widget_from_tree(self.settings_tree, TableViewCustom)[0]
        self.settings.child('tabular_table').setValue(self.table_model)
        self.n_axes = len(self._actuators)
        self.update_table_view()

    def update_table_view(self):
        self.table_view.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.table_view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        styledItemDelegate = QtWidgets.QStyledItemDelegate()
        styledItemDelegate.setItemEditorFactory(gutils.SpinBoxDelegate())
        self.table_view.setItemDelegate(styledItemDelegate)

        self.table_view.setDragEnabled(True)
        self.table_view.setDropIndicatorShown(True)
        self.table_view.setAcceptDrops(True)
        self.table_view.viewport().setAcceptDrops(True)
        self.table_view.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.table_view.setDragDropMode(QtWidgets.QTableView.InternalMove)
        self.table_view.setDragDropOverwriteMode(False)

        self.table_view.add_data_signal[int].connect(self.table_model.add_data)
        self.table_view.remove_row_signal[int].connect(self.table_model.remove_data)
        self.table_view.load_data_signal.connect(self.table_model.load_txt)
        self.table_view.save_data_signal.connect(self.table_model.save_txt)

    def evaluate_steps(self):
        return len(self.table_model)

    def set_scan(self):
        positions = np.array(self.table_model.get_data_all())
        self.get_info_from_positions(positions)

    def update_tabular_positions(self, positions: np.ndarray = None):
        """Convenience function to write positions directly into the tabular table

        Parameters
        ----------
        positions: ndarray
            a 2D ndarray with as many columns as selected actuators
        """
        try:
            if positions is None:
                if self.settings.child('tabular_settings',
                                       'tabular_selection').value() == 'Polylines':  # from ROI
                    viewer = self.scan_selector.selector_source

                    if self.settings.child('tabular_settings', 'tabular_subtype').value() == 'Linear':
                        positions = self.scan_selector.scan_selector.getArrayIndexes(
                            spacing=self.settings.child('tabular_settings', 'tabular_step').value())
                    elif self.settings.child('tabular_settings',
                                             'tabular_subtype').value() == 'Adaptive':
                        positions = self.scan_selector.scan_selector.get_vertex()

                    steps_x, steps_y = zip(*positions)
                    steps_x, steps_y = viewer.scale_axis(np.array(steps_x), np.array(steps_y))
                    positions = np.transpose(np.array([steps_x, steps_y]))
                    self.update_model(init_data=positions)
                else:
                    self.update_model()
            elif isinstance(positions, np.ndarray):
                self.update_model(init_data=positions)
            else:
                pass
        except Exception as e:
            logger.exception(str(e))

    def get_nav_axes(self) -> List[Axis]:
        return [Axis(label=f'{act.title}', units=act.units, data=self.positions[:, ind], index=0,
                     spread_order=ind)
                for ind, act in enumerate(self.actuators)]

    def get_indexes_from_scan_index(self, scan_index: int) -> Tuple[int]:
        """To be reimplemented. Calculations of indexes within the scan"""
        return scan_index,

    def get_scan_shape(self) -> Tuple[int]:
        return len(self.table_model),

    def update_from_scan_selector(self, scan_selector: Selector):
        coordinates = scan_selector.get_coordinates()
        self.update_model(init_data=coordinates)


@ScannerFactory.register()
class TabularScannerSubsegmented(TabularScanner):
    scan_subtype = 'Subsegmented'
    save_settings = False
    params = [{'title': 'Step:', 'name': 'tabular_step', 'type': 'float', 'value': 0.1},
              {'title': 'Points', 'name': 'tabular_points', 'type': 'table_view', 'delegate': gutils.SpinBoxDelegate,
               'menu': True},
             ] + TabularScanner.params

    def __init__(self, actuators: List[str]):
        self.table_model: TableModelTabularReadOnly = None
        self.table_view: TableViewCustom = None
        self.table_model_points: TableModelTabular = None
        self.table_view_points: TableViewCustom = None
        super().__init__(actuators=actuators)
        self.update_model()
        self.update_model_points()

    @property
    def actuators(self):
        return self._actuators

    @actuators.setter
    def actuators(self, actuators_name):
        self._actuators = actuators_name
        self.update_model()
        self.update_model_points()

    def update_model(self, init_data=None):
        if init_data is None:
            init_data = [[0. for _ in self._actuators]]

        self.table_model = TableModelTabularReadOnly(init_data, [act.title for act in self._actuators])
        self.table_view = putils.get_widget_from_tree(self.settings_tree, TableViewCustom)[1]
        self.settings.child('tabular_table').setValue(self.table_model)
        self.n_axes = len(self._actuators)
        self.update_table_view()

    def update_model_points(self, init_data=None):
        if init_data is None:
            init_data = [[0. for _ in self._actuators]]

        self.table_model_points = TableModelTabular(init_data, [act.title for act in self._actuators])
        self.table_view_points = putils.get_widget_from_tree(self.settings_tree, TableViewCustom)[0]
        self.settings.child('tabular_points').setValue(self.table_model_points)
        self.update_table_view_points()

    def update_table_view(self):
        self.table_view.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.table_view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        # self.table_view.setEnabled(False)

    def update_table_view_points(self):
        self.table_view_points.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table_view_points.horizontalHeader().setStretchLastSection(True)
        self.table_view_points.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.table_view_points.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        styledItemDelegate = QtWidgets.QStyledItemDelegate()
        styledItemDelegate.setItemEditorFactory(gutils.SpinBoxDelegate())
        self.table_view.setItemDelegate(styledItemDelegate)

        self.table_view_points.setDragEnabled(True)
        self.table_view_points.setDropIndicatorShown(True)
        self.table_view_points.setAcceptDrops(True)
        self.table_view_points.viewport().setAcceptDrops(True)
        self.table_view_points.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.table_view_points.setDragDropMode(QtWidgets.QTableView.InternalMove)
        self.table_view_points.setDragDropOverwriteMode(False)

        self.table_view_points.add_data_signal[int].connect(self.table_model_points.add_data)
        self.table_view_points.remove_row_signal[int].connect(self.table_model_points.remove_data)
        self.table_view_points.load_data_signal.connect(self.table_model_points.load_txt)
        self.table_view_points.save_data_signal.connect(self.table_model_points.save_txt)

    def set_scan(self):
        points = [Point(coordinates) for coordinates in self.table_model_points.get_data_all()]
        positions = get_sub_segmented_positions(self.settings['tabular_step'], points)

        self.table_model.set_data_all(positions)
        positions = np.array(self.table_model.get_data_all())
        self.get_info_from_positions(positions)

    def update_from_scan_selector(self, scan_selector: Selector):
        coordinates = scan_selector.get_coordinates()
        self.update_model_points(init_data=coordinates)
        self.set_scan()


# -*- coding: utf-8 -*-
"""
Created the 20/10/2023

@author: Sebastien Weber
"""
import numbers
from typing import Tuple, List, Union, TYPE_CHECKING


import numpy as np
from pymodaq.utils import data as data_mod
from pymodaq.utils.data import DataWithAxes, DataToExport, Axis, DwaType

if TYPE_CHECKING:
    from pymodaq.utils.tcp_ip.mysocket import Socket


class SocketString:
    """Mimic the Socket object but actually using a bytes string not a socket connection

    Implements a minimal interface of two methods

    Parameters
    ----------
    bytes_string: bytes

    See Also
    --------
    :class:`~pymodaq.utils.tcp_ip.mysocket.Socket`
    """
    def __init__(self, bytes_string: bytes):
        self._bytes_string = bytes_string

    def check_received_length(self, length: int) -> bytes:
        """
        Make sure all bytes (length) that should be received are received through the socket.

        Here just read the content of the underlying bytes string

        Parameters
        ----------
        length: int
            The number of bytes to be read from the socket

        Returns
        -------
        bytes
        """
        data = self._bytes_string[0:length]
        self._bytes_string = self._bytes_string[length:]
        return data

    def get_first_nbytes(self, length: int) -> bytes:
        """ Read the first N bytes from the socket

        Parameters
        ----------
        length: int
            The number of bytes to be read from the socket

        Returns
        -------
        bytes
            the read bytes string
        """
        return self.check_received_length(length)


class Serializer:
    """Used to Serialize to bytes python objects, numpy arrays and PyMoDAQ DataWithAxes and DataToExport objects"""

    def __init__(self, obj: Union[int, str, numbers.Number, list, np.ndarray, Axis, DataWithAxes, DataToExport] = None):
        self._bytes_string = b''
        self._obj = obj

    def to_bytes(self):
        """ Generic method to obtain the bytes string from various objects

        Compatible objects are:

        * :class:`bytes`
        * :class:`numbers.Number`
        * :class:`str`
        * :class:`numpy.ndarray`
        * :class:`~pymodaq.utils.data.Axis`
        * :class:`~pymodaq.utils.data.DataWithAxes` and sub-flavours
        * :class:`~pymodaq.utils.data.DataToExport`
        * :class:`list` of any objects above

        """
        if isinstance(self._obj, bytes):
            return self.bytes_serialization(self._obj)
        elif isinstance(self._obj, numbers.Number):
            return self.scalar_serialization(self._obj)
        elif isinstance(self._obj, str):
            return self.string_serialization(self._obj)
        elif isinstance(self._obj, np.ndarray):
            return self.ndarray_serialization(self._obj)
        elif isinstance(self._obj, Axis):
            return self.axis_serialization(self._obj)
        elif self._obj.__class__.__name__ in DwaType.names():
            return self.dwa_serialization(self._obj)
        elif isinstance(self._obj, DataToExport):
            return self.dte_serialization(self._obj)
        elif isinstance(self._obj, list):
            return self.list_serialization(self._obj)

    @staticmethod
    def int_to_bytes(an_integer: int) -> bytes:
        """Convert an unsigned integer into a byte array of length 4 in big endian

        Parameters
        ----------
        an_integer: int

        Returns
        -------
        bytearray
        """
        if not isinstance(an_integer, int):
            raise TypeError(f'{an_integer} should be an integer, not a {type(an_integer)}')
        elif an_integer < 0:
            raise ValueError(f'Can only serialize unsigned integer using this method')
        return an_integer.to_bytes(4, 'big')

    @staticmethod
    def str_to_bytes(message: str) -> bytes:
        if not isinstance(message, str):
            raise TypeError(f'Can only serialize str object using this method')
        return message.encode()

    @classmethod
    def str_len_to_bytes(cls, message: str) -> (bytes, bytes):
        """ Convert a string and its length to two bytes
        Parameters
        ----------
        message: str
            the message to convert

        Returns
        -------
        bytes: message converted as a byte array
        bytes: length of the message byte array, itself as a byte array of length 4
        """

        if not isinstance(message, str) and not isinstance(message, bytes):
            message = str(message)
        if not isinstance(message, bytes):
            message = cls.str_to_bytes(message)
        return message, cls.int_to_bytes(len(message))

    def _int_serialization(self, int_obj: int) -> bytes:
        """serialize an unsigned integer used for getting the length of messages internaly, for outside integer
        serialization or deserialization use scalar_serialization"""
        int_bytes = self.int_to_bytes(int_obj)
        bytes_string = int_bytes
        self._bytes_string += bytes_string
        return bytes_string

    def bytes_serialization(self, bytes_string_in: bytes) -> bytes:
        bytes_string = b''
        bytes_string += self.int_to_bytes(len(bytes_string_in))
        bytes_string += bytes_string_in
        return bytes_string

    def string_serialization(self, string: str) -> bytes:
        """ Convert a string into a bytes message together with the info to convert it back

        Parameters
        ----------
        string: str

        Returns
        -------
        bytes: the total bytes message to serialize the string
        """
        bytes_string = b''
        cmd_bytes, cmd_length_bytes = self.str_len_to_bytes(string)
        bytes_string += cmd_length_bytes
        bytes_string += cmd_bytes
        self._bytes_string += bytes_string
        return bytes_string

    def scalar_serialization(self, scalar: numbers.Number) -> bytes:
        """ Convert a scalar into a bytes message together with the info to convert it back

        Parameters
        ----------
        scalar: str

        Returns
        -------
        bytes: the total bytes message to serialize the scalar
        """
        if not isinstance(scalar, numbers.Number):
            raise TypeError(f'{scalar} should be an integer or a float, not a {type(scalar)}')
        scalar_array = np.array([scalar])
        data_type = scalar_array.dtype.descr[0][1]
        data_bytes = scalar_array.tobytes()

        bytes_string = b''
        bytes_string += self.string_serialization(data_type)
        bytes_string += self._int_serialization(len(data_bytes))
        bytes_string += data_bytes
        self._bytes_string += bytes_string
        return bytes_string

    def ndarray_serialization(self, array: np.ndarray) -> bytes:
        """ Convert a ndarray into a bytes message together with the info to convert it back

        Parameters
        ----------
        array: np.ndarray

        Returns
        -------
        bytes: the total bytes message to serialize the scalar

        Notes
        -----

        The bytes sequence is constructed as:

        * get data type as a string
        * reshape array as 1D array and get the array dimensionality (len of array's shape)
        * convert Data array as bytes
        * serialize data type
        * serialize data length
        * serialize data shape length
        * serialize all values of the shape as integers converted to bytes
        * serialize array as bytes
        """
        if not isinstance(array, np.ndarray):
            raise TypeError(f'{array} should be an numpy array, not a {type(array)}')
        array_type = array.dtype.descr[0][1]
        array_shape = array.shape

        array = array.reshape(array.size)
        array_bytes = array.tobytes()
        bytes_string = b''
        bytes_string += self.string_serialization(array_type)
        bytes_string += self._int_serialization(len(array_bytes))
        bytes_string += self._int_serialization(len(array_shape))
        for shape_elt in array_shape:
            bytes_string += self._int_serialization(shape_elt)
        bytes_string += array_bytes
        self._bytes_string += bytes_string
        return bytes_string

    def object_type_serialization(self, obj: Union[Axis, DataToExport, DataWithAxes]) -> bytes:
        """ Convert an object type into a bytes message as a string together with the info to convert it back

        Applies to Data object from the pymodaq.utils.data module
        """
        return self.string_serialization(obj.__class__.__name__)

    def axis_serialization(self, axis: Axis) -> bytes:
        """ Convert an Axis object into a bytes message together with the info to convert it back

        Parameters
        ----------
        axis: Axis

        Returns
        -------
        bytes: the total bytes message to serialize the Axis

        Notes
        -----

        The bytes sequence is constructed as:

        * serialize the type: 'Axis'
        * serialize the axis label
        * serialize the axis units
        * serialize the axis array
        * serialize the axis
        * serialize the axis spread_order
        """
        if not isinstance(axis, Axis):
            raise TypeError(f'{axis} should be a list, not a {type(axis)}')

        bytes_string = b''
        bytes_string += self.object_type_serialization(axis)
        bytes_string += self.string_serialization(axis.label)
        bytes_string += self.string_serialization(axis.units)
        bytes_string += self.ndarray_serialization(axis.get_data())
        bytes_string += self.scalar_serialization(axis.index)
        bytes_string += self.scalar_serialization(axis.spread_order)
        self._bytes_string += bytes_string
        return bytes_string

    def list_serialization(self, list_object: List) -> bytes:
        """ Convert a list of objects into a bytes message together with the info to convert it back

        Parameters
        ----------
        list_object: list
            the list could contains either scalars, strings or ndarrays or Axis objects or DataWithAxis objects
            module

        Returns
        -------
        bytes: the total bytes message to serialize the list of objects

        Notes
        -----

        The bytes sequence is constructed as:
        * the length of the list

        Then for each object:

        * get data type as a string
        * use the serialization method adapted to each object in the list
        """
        if not isinstance(list_object, list):
            raise TypeError(f'{list_object} should be a list, not a {type(list_object)}')

        bytes_string = b''

        bytes_string += self._int_serialization(len(list_object))
        for obj in list_object:
            if isinstance(obj, DataWithAxes):
                bytes_string += self.string_serialization('dwa')
                bytes_string += self.dwa_serialization(obj)

            elif isinstance(obj, Axis):
                bytes_string += self.string_serialization('axis')
                bytes_string += self.axis_serialization(obj)

            elif isinstance(obj, np.ndarray):
                bytes_string += self.string_serialization('array')
                bytes_string += self.ndarray_serialization(obj)

            elif isinstance(obj, str):
                bytes_string += self.string_serialization('string')
                bytes_string += self.string_serialization(obj)

            elif isinstance(obj, numbers.Number):
                bytes_string += self.string_serialization('scalar')
                bytes_string += self.scalar_serialization(obj)

            else:
                raise TypeError(f'the element {obj} type cannot be serialized into bytes, only numpy arrays'
                                f', strings, or scalars (int or float)')
        self._bytes_string += bytes_string
        return bytes_string

    def dwa_serialization(self, dwa: DataWithAxes) -> bytes:
        """ Convert a DataWithAxes into a bytes string

        Parameters
        ----------
        dwa: DataWithAxes

        Returns
        -------
        bytes: the total bytes message to serialize the DataWithAxes

        Notes
        -----
        The bytes sequence is constructed as:

        * serialize the string type: 'DataWithAxes'
        * serialize the timestamp: float
        * serialize the name
        * serialize the source enum as a string
        * serialize the dim enum as a string
        * serialize the distribution enum as a string
        * serialize the list of numpy arrays
        * serialize the list of labels
        * serialize the origin
        * serialize the nav_index tuple as a list of int
        * serialize the list of axis
        """
        if not isinstance(dwa, DataWithAxes):
            raise TypeError(f'{dwa} should be a DataWithAxes, not a {type(dwa)}')

        bytes_string = b''
        bytes_string += self.object_type_serialization(dwa)
        bytes_string += self.scalar_serialization(dwa.timestamp)
        bytes_string += self.string_serialization(dwa.name)
        bytes_string += self.string_serialization(dwa.source.name)
        bytes_string += self.string_serialization(dwa.dim.name)
        bytes_string += self.string_serialization(dwa.distribution.name)
        bytes_string += self.list_serialization(dwa.data)
        bytes_string += self.list_serialization(dwa.labels)
        bytes_string += self.string_serialization(dwa.origin)
        bytes_string += self.list_serialization(list(dwa.nav_indexes))
        bytes_string += self.list_serialization(dwa.axes)
        self._bytes_string += bytes_string
        return bytes_string

    def dte_serialization(self, dte: DataToExport) -> bytes:
        """ Convert a DataToExport into a bytes string

        Parameters
        ----------
        dte: DataToExport

        Returns
        -------
        bytes: the total bytes message to serialize the DataToExport

        Notes
        -----
        The bytes sequence is constructed as:

        * serialize the string type: 'DataToExport'
        * serialize the timestamp: float
        * serialize the name
        * serialize the list of DataWithAxes
        """
        if not isinstance(dte, DataToExport):
            raise TypeError(f'{dte} should be a DataToExport, not a {type(dte)}')

        bytes_string = b''
        bytes_string += self.object_type_serialization(dte)
        bytes_string += self.scalar_serialization(dte.timestamp)
        bytes_string += self.string_serialization(dte.name)
        bytes_string += self.list_serialization(dte.data)
        self._bytes_string += bytes_string
        return bytes_string


class DeSerializer:
    """Used to DeSerialize bytes to python objects, numpy arrays and PyMoDAQ Axis, DataWithAxes and DataToExport
    objects

    Parameters
    ----------
    bytes_string: bytes or Socket
        the bytes string to deserialize into an object: int, float, string, arrays, list, Axis, DataWithAxes...
        Could also be a Socket object reading bytes from the network having a `get_first_nbytes` method

    See Also
    --------
    :py:class:`~pymodaq.utils.tcp_ip.serializer.SocketString`
    :py:class:`~pymodaq.utils.tcp_ip.mysocket.Socket`
    """

    def __init__(self, bytes_string:  Union[bytes, 'Socket'] = None):
        if isinstance(bytes_string, bytes):
            bytes_string = SocketString(bytes_string)
        self._bytes_string = bytes_string

    @staticmethod
    def bytes_to_string(message: bytes) -> str:
        return message.decode()

    @staticmethod
    def bytes_to_int(bytes_string: bytes) -> int:
        """Convert a bytes of length 4 into an integer"""
        if not isinstance(bytes_string, bytes):
            raise TypeError(f'{bytes_string} should be an bytes string, not a {type(bytes_string)}')
        assert len(bytes_string) == 4
        return int.from_bytes(bytes_string, 'big')

    @staticmethod
    def bytes_to_scalar(data: bytes, dtype: np.dtype) -> numbers.Number:
        """Convert bytes to a scalar given a certain numpy dtype

        Parameters
        ----------
        data: bytes
        dtype:np.dtype

        Returns
        -------
        numbers.Number
        """
        return np.frombuffer(data, dtype=dtype)[0]

    @staticmethod
    def bytes_to_nd_array(data: bytes, dtype: np.dtype, shape: Tuple[int]) -> np.ndarray:
        """Convert bytes to a ndarray given a certain numpy dtype and shape

        Parameters
        ----------
        data: bytes
        dtype: np.dtype
        shape: tuple of int

        Returns
        -------
        np.ndarray
        """
        array = np.frombuffer(data, dtype=dtype)
        array = array.reshape(tuple(shape))
        array = np.atleast_1d(array)  # remove singleton dimensions but keeping ndarrays
        return array

    def _int_deserialization(self) -> int:
        """Convert the fourth first bytes into an unsigned integer to be used internally. For integer serialization
        use scal_serialization"""
        int_obj = self.bytes_to_int(self._bytes_string.get_first_nbytes(4))
        return int_obj

    def string_deserialization(self) -> str:
        """Convert bytes into a str object

        Convert first the fourth first bytes into an int encoding the length of the string to decode

        Returns
        -------
        str: the decoded string
        """
        string_len = self._int_deserialization()
        str_obj = self._bytes_string.get_first_nbytes(string_len).decode()
        return str_obj

    def scalar_deserialization(self) -> numbers.Number:
        """Convert bytes into a numbers.Number object

        Get first the data type from a string deserialization, then the data length and finally convert this
        length of bytes into a number (float, int)

        Returns
        -------
        numbers.Number: the decoded number
        """
        data_type = self.string_deserialization()
        data_len = self._int_deserialization()
        number = np.frombuffer(self._bytes_string.get_first_nbytes(data_len), dtype=data_type)[0]
        if 'f' in data_type:
            number = float(number)  # because one get numpy  float type
        elif 'i' in data_type:
            number = int(number)  # because one get numpy int type
        return number

    def ndarray_deserialization(self) -> np.ndarray:
        """Convert bytes into a numpy ndarray object

        Convert the first bytes into a ndarray reading first information about the array's data

        Returns
        -------
        ndarray: the decoded numpy array
        """
        ndarray_type = self.string_deserialization()
        ndarray_len = self._int_deserialization()
        shape_len = self._int_deserialization()
        shape = []
        for ind in range(shape_len):
            shape_elt = self._int_deserialization()
            shape.append(shape_elt)

        ndarray = np.frombuffer(self._bytes_string.get_first_nbytes(ndarray_len), dtype=ndarray_type)
        ndarray = ndarray.reshape(tuple(shape))
        ndarray = np.atleast_1d(ndarray)  # remove singleton dimensions
        return ndarray

    def list_deserialization(self) -> list:
        """Convert bytes into a list of homogeneous objects

        Convert the first bytes into a list reading first information about the list elt types, length ...

        Returns
        -------
        list: the decoded list
        """
        list_obj = []
        list_len = self._int_deserialization()

        for ind in range(list_len):
            obj_type = self.string_deserialization()
            if obj_type == 'scalar':
                list_elt = self.scalar_deserialization()
            elif obj_type == 'string':
                list_elt = self.string_deserialization()
            elif obj_type == 'array':
                list_elt = self.ndarray_deserialization()
            elif obj_type == 'dwa':
                list_elt = self.dwa_deserialization()
            elif obj_type == 'axis':
                list_elt = self.axis_deserialization()
            list_obj.append(list_elt)
        return list_obj

    def axis_deserialization(self) -> Axis:
        """Convert bytes into an Axis object

        Convert the first bytes into an Axis reading first information about the Axis

        Returns
        -------
        Axis: the decoded Axis
        """

        class_name = self.string_deserialization()
        if class_name != Axis.__name__:
            raise TypeError(f'Attempting to deserialize an Axis but got the bytes for a {class_name}')
        axis_label = self.string_deserialization()
        axis_units = self.string_deserialization()
        axis_array = self.ndarray_deserialization()
        axis_index = self.scalar_deserialization()
        axis_spread_order = self.scalar_deserialization()

        axis = Axis(axis_label, axis_units, data=axis_array, index=axis_index, spread_order=axis_spread_order)
        return axis

    def dwa_deserialization(self) -> DataWithAxes:
        """Convert bytes into a DataWithAxes object

        Convert the first bytes into a DataWithAxes reading first information about the underlying data

        Returns
        -------
        DataWithAxes: the decoded DataWithAxes
        """
        class_name = self.string_deserialization()
        if class_name not in DwaType.names():
            raise TypeError(f'Attempting to deserialize a DataWithAxes flavor but got the bytes for a {class_name}')
        timestamp = self.scalar_deserialization()
        dwa = getattr(data_mod, class_name)(self.string_deserialization(),
                                            source=self.string_deserialization(),
                                            dim=self.string_deserialization(),
                                            distribution=self.string_deserialization(),
                                            data=self.list_deserialization(),
                                            labels=self.list_deserialization(),
                                            origin=self.string_deserialization(),
                                            nav_indexes=tuple(self.list_deserialization()),
                                            axes=self.list_deserialization(),
                                            )
        dwa.timestamp = timestamp
        return dwa

    def dte_deserialization(self) -> DataToExport:
        """Convert bytes into a DataToExport object

        Convert the first bytes into a DataToExport reading first information about the underlying data

        Returns
        -------
        DataToExport: the decoded DataToExport
        """
        class_name = self.string_deserialization()
        if class_name != DataToExport.__name__:
            raise TypeError(f'Attempting to deserialize a DataToExport but got the bytes for a {class_name}')
        timestamp = self.scalar_deserialization()
        dte = DataToExport(self.string_deserialization(),
                           data=self.list_deserialization(),
                           )
        dte.timestamp = timestamp
        return dte


# -*- coding: utf-8 -*-
"""
Created the 02/03/2023

@author: Nicolas Tappy
"""
# Standard imports

# 3rd party imports
try:
    import numpy as np
    import hyperspy.api_nogui as hs

    # project imports
    from pymodaq.utils.h5modules.backends import Node
    from pymodaq.utils.h5modules.utils import get_h5_data_from_node, extract_axis, verify_axis_data_uniformity
    from pymodaq.utils.h5modules.exporter import ExporterFactory, H5Exporter
    from pymodaq.utils.logger import set_logger, get_module_name
    from pymodaq.utils.h5modules.data_saving import DataLoader

    # This is needed to log
    logger = set_logger(get_module_name(__file__))


    @ExporterFactory.register_exporter()
    class H5hspyExporter(H5Exporter):
        """ Exporter object for saving nodes as hspy files"""

        FORMAT_DESCRIPTION = "Hyperspy file format"
        FORMAT_EXTENSION = "hspy"

        def export_data(self, node: Node, filename) -> None:
            """Exporting a .h5 node as a hyperspy object"""

            # first verify if the node type is compatible with export. Only data nodes are.
            ############## compatible with v4 done by Seb
            nodetype = node.attrs['data_type']
            if 'data' in nodetype.lower():
                # If yes we can use this function (the same for plotting in h5browser) to extract information

                data_loader = DataLoader(node.to_h5_backend())

                #use the method in data_saving below to load properly data from nodes. The data object is a DataWithAxes
                #of dimensionality DataND
                #you can get all axes (nav and signals) from data.axes
                #each axis as an index corresponding to the data shape
                # data.nav_index is a tuple containing the index of the nav axes

                data = data_loader.load_data(node, with_bkg=False)
                axes = data.axes #all axes
                nav_axes = data.get_nav_axes_with_data() #only navigation one

                #deprecated
                # data, axes, nav_axes, is_spread = get_h5_data_from_node(node)
                is_spread = node.attrs['distribution'] != 'uniform'
                
                data_array = data.data.reshape(data.shape, order='F')
                logger.debug(f"extracted data of shape {data.shape}")
                logger.debug(f"extracted axes: {axes}")
                #################################### TODO below to be done by Nicolas :-) and should be tested!!

                # get_h5_data_from_node will return sometimes empty axes that we must yeet.
                axes_keys_to_remove = []
                for key, ax in axes.items():
                    if ax['data'] is None:
                        logger.warning(f'Dropping axis {key} due to empty data')
                        axes_keys_to_remove.append(key)
                for key in axes_keys_to_remove:
                    del axes[key]

            else:
                # Else just raise an error.
                raise NotImplementedError(f"hspy export not supported for nodes of type {nodetype}'. "
                                          f"Export from the data node instead.")

            # Verify that this is not an adaptive scan. If it is we throw and say it is not supported.
            if is_spread:
                raise NotImplementedError(f"hspy export not supported for adaptive scan data.")

            # Then we build the hyperspy axes objects.
            hyperspy_axes = []
            for key, ax in axes.items():
                logger.info(f"Extracting axis {ax['label']}.")
                unique_ax = extract_axis(ax)
                try:
                    indexv = data.shape.index(len(unique_ax))
                    logger.info(f"{ax['label']}, size {len(unique_ax)}, data index: {indexv}")
                except ValueError:
                    raise ValueError(f"Signal axis of length {len(unique_ax)} does not match any "
                                     f"dimension in data of shape {data.shape} ")
                is_nav = key.startswith('nav')
                hyperspy_axes.append(self.build_hyperspy_axis(unique_ax, data_idx=indexv, label=ax['label'],
                                                              unit=ax['units'], navigate=is_nav))

            ordered_axes = sorted(hyperspy_axes, key=lambda d: d['index_in_array'])
            for ax in ordered_axes:
                del ax['index_in_array']
            # Then we build the hyperspy object. First we must know its dimensionality
            # from the number of signal axes.
            dim = len(axes) - len(nav_axes)
            if dim == 1:
                # Then signal1D
                sig = hs.signals.Signal1D(data=data, original_metadata={}, axes=ordered_axes)
            elif dim == 2:
                # Then signal2D
                sig = hs.signals.Signal2D(data=data, original_metadata={}, axes=ordered_axes)
            else:
                # Then basesignal
                sig = hs.signals.BaseSignal(data=data, original_metadata={}, axes=ordered_axes)

            # Finally save
            sig.save(filename)

        def build_hyperspy_axis(self, ax_data: np.ndarray, data_idx: int,
                                label: str, unit: str, navigate: bool) -> dict:
            """Build an axis based on the input data. Choose between a UniformDataAxis or
            DataAxis object based on a quick linearity check of the input data."""
            offset, scale = verify_axis_data_uniformity(ax_data)
            if offset is not None:
                axis_dict = {'_type': 'UniformDataAxis',
                             'index_in_array': data_idx,
                             'name': label,
                             'units': unit,
                             'navigate': navigate,
                             'size': len(ax_data),
                             'scale': scale,
                             'offset': offset}
            else:
                axis_dict = {'_type': 'DataAxis',
                             'index_in_array': data_idx,
                             'name': label,
                             'units': unit,
                             'navigate': navigate,
                             'size': len(ax_data),
                             'axis': ax_data}

            return axis_dict

        def build_hyperspy_original_metadata(self, node: Node):
            """Build original metadata dictionary"""
            pass
except:
    pass



from pathlib import Path

from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtWidgets import QWidget
from pymodaq.utils.gui_utils.widgets.spinbox import SpinBox


EDIT_PUSH_TYPES = ['abs', 'rel']


class PushButtonIcon(QtWidgets.QPushButton):
    def __init__(self, icon_name: str, text: str, checkable=False, tip="", menu=None):
        super().__init__(text)
        self._menu = menu
        if icon_name != '':
            icon = QtGui.QIcon()
            if Path(icon_name).is_file():
                icon.addPixmap(QtGui.QPixmap(icon_name), QtGui.QIcon.Normal,
                               QtGui.QIcon.Off)
            else:
                icon.addPixmap(QtGui.QPixmap(f":/icons/Icon_Library/{icon_name}.png"), QtGui.QIcon.Normal,
                               QtGui.QIcon.Off)
            self.setIcon(icon)
        self.setCheckable(checkable)
        self.setToolTip(tip)
        
    def contextMenuEvent(self, event):
        if self._menu is not None:
            self._menu.exec(event.globalPos())


class EditPushInfo:
    def __init__(self, type:str, value: float):
        if type not in EDIT_PUSH_TYPES:
            raise ValueError(f'Possible types are {EDIT_PUSH_TYPES}')
        self.type = type
        self.value = value


class EditPush(QtWidgets.QWidget):
    clicked = QtCore.Signal(EditPushInfo)

    def __init__(self, icon_name: str, ini_value=0.1, text=''):
        super().__init__()
        self._edit_push_type = 'abs'
        self.setLayout(QtWidgets.QHBoxLayout())
        self._edit = SpinBox(value=ini_value, dec=True, step=0.1, minStep=0.001)
        self._edit.setMinimumHeight(20)
        self._edit.setMaximumWidth(60)
        self._edit.setMinimumWidth(60)
        self.layout().addWidget(self._edit)
        self.set_pushs(icon_name, text)

    def set_pushs(self, icon_name, text):
        self._push = PushButtonIcon(icon_name, text)
        self._push.setMaximumWidth(40)
        self._push.clicked.connect(lambda: self.emit_clicked())
        self.layout().addWidget(self._push)

    def emit_clicked(self, coeff=1):
        """will emit a signal containing a float value calculated from the product of the coeff and the internal
        spinbox value.

        See Also
        --------
        EditPushRel
        """
        self.clicked.emit(EditPushInfo(type=self._edit_push_type, value=coeff * self._edit.value()))


class EditPushRel(EditPush):

    def __init__(self, icon_name: str, text='', ini_value=0.15):
        super().__init__(icon_name, text=text, ini_value=ini_value)
        self._edit_push_type = 'rel'

    def set_pushs(self, icon_name, text):
        vlayout = QtWidgets.QVBoxLayout()
        self.layout().addLayout(vlayout)

        self._push_plus = PushButtonIcon(icon_name, f'+{text}')
        self._push_plus.setMaximumWidth(40)
        self._push_minus = PushButtonIcon(icon_name, f'-{text}')
        self._push_minus.setMaximumWidth(40)

        vlayout.addWidget(self._push_plus)
        vlayout.addWidget(self._push_minus)

        self._push_plus.clicked.connect(lambda: self.emit_clicked(1))
        self._push_minus.clicked.connect(lambda: self.emit_clicked(-1))


class ActionMenu(QtWidgets.QAction):
    def __init__(self, *args, menu=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._menu = menu

    def contextMenuEvent(self, event):
        if self._menu is not None:
            self._menu.exec(event.globalPos())


def main(init_qt=True):
    import sys

    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)

    widget = QtWidgets.QWidget()
    widget.setLayout(QtWidgets.QVBoxLayout())
    toolbar = QtWidgets.QToolBar()
    widget.layout().addWidget(toolbar)

    action = QtWidgets.QAction('blabla')
    toolbar.addAction(action)

    menu1 = QtWidgets.QMenu()
    menu1.addAction('Add new task')
    menu1.addAction('Edit current row')
    menu1.addAction('Remove selected rows')
    menu1.addSeparator()
    menu1.addAction('Affect Responsable')
    menu2 = QtWidgets.QMenu()
    menu2.addAction('Affect volunteers')
    menu2.addAction('Remove volunteers')
    menu2.addSeparator()
    menu2.addAction('Show localisation')

    icon = QtGui.QIcon()
    icon.addPixmap(QtGui.QPixmap(f":/icons/Icon_Library/{'run2'}.png"), QtGui.QIcon.Normal,
                   QtGui.QIcon.Off)
    action2 = ActionMenu(icon, 'grab', menu=menu1)

    toolbar.addAction(action2)
    grab_pb = PushButtonIcon('run2', text='', checkable=True, menu=menu2)
    grab_pb.setFlat(True)
    toolbar.addWidget(grab_pb)

    otherpb = PushButtonIcon('run2', 'grab me')
    widget.layout().addWidget(otherpb)

    widget.show()
    if init_qt:
        sys.exit(app.exec_())


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
"""
Created the 23/11/2022

@author: Sebastien Weber
"""
from __future__ import annotations

from typing import Union, List, Dict, Tuple, TYPE_CHECKING
import xml.etree.ElementTree as ET


import numpy as np

from pymodaq.utils.abstract import ABCMeta, abstract_attribute, abstractmethod
from pymodaq.utils.daq_utils import capitalize
from pymodaq.utils.data import Axis, DataDim, DataWithAxes, DataToExport, DataDistribution
from .saving import H5SaverLowLevel
from .backends import GROUP, CARRAY, Node, GroupType
from .data_saving import DataToExportSaver, AxisSaverLoader, DataToExportTimedSaver, DataToExportExtendedSaver
from pymodaq.utils.parameter import ioxml

if TYPE_CHECKING:
    from pymodaq.extensions.daq_scan import DAQScan
    from pymodaq.control_modules.daq_viewer import DAQ_Viewer
    from pymodaq.control_modules.daq_move import DAQ_Move
    from pymodaq.utils.h5modules.h5logging import H5Logger


class ModuleSaver(metaclass=ABCMeta):
    """Abstract base class to save info and data from main modules (DAQScan, DAQViewer, DAQMove, ...)"""
    group_type: GroupType = abstract_attribute()
    _module = abstract_attribute()
    _h5saver: H5SaverLowLevel = abstract_attribute()
    _module_group: GROUP = abstract_attribute()
    main_module = True

    def flush(self):
        """Flush the underlying file"""
        self._h5saver.flush()

    def get_set_node(self, where: Union[Node, str] = None, name: str = None) -> GROUP:
        """Get or create the node corresponding to this particular Module instance

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        new: bool
            if True force the creation of a new indexed node of this class type
            if False return the last node (or create one if None)

        Returns
        -------
        GROUP: the Node associated with this module which should be a GROUP node
        """
        if where is None:
            where = self._h5saver.raw_group
        if name is None:
            name = self._module.title
        group = self._h5saver.get_node_from_title(where, name)
        if group is not None:
            self._module_group = group
            return group  # if I got one I return it else I create one

        self._module_group = self._add_module(where)
        return self._module_group

    def get_last_node(self, where: Union[Node, str] = None):
        """Get the last node corresponding to this particular Module instance

        Parameters
        ----------
        where: Union[Node, str]
           the path of a given node or the node itself
        new: bool
           if True force the creation of a new indexed node of this class type
           if False return the last node (or create one if None)

        Returns
        -------
        GROUP: the Node associated with this module which should be a GROUP node
        """
        if where is None:
            where = self._h5saver.raw_group

        group = self._h5saver.get_last_group(where, self.group_type)
        self._module_group = group
        return self._module_group

    @abstractmethod
    def _add_module(self, where: Union[Node, str] = None, metadata={}):
        ...

    @property
    def module(self):
        return self._module

    @property
    def module_group(self):
        return self._module_group

    @property
    def h5saver(self):
        return self._h5saver

    @h5saver.setter
    def h5saver(self, _h5saver: H5SaverLowLevel):
        self._h5saver = _h5saver
        self.update_after_h5changed()

    @abstractmethod
    def update_after_h5changed(self):
        ...

    def get_last_node_index(self, where: Union[Node, str] = None):
        node = self.get_last_node(where)
        return int(node.name.split(capitalize(self.group_type.name))[1])

    def get_next_node_name(self, where: Union[Node, str] = None):
        index = self.get_last_node_index(where)
        return f'{capitalize(self.group_type.name)}{index+1:03d}'


class DetectorSaver(ModuleSaver):
    """Implementation of the ModuleSaver class dedicated to DAQ_Viewer modules

    Parameters
    ----------
    module
    """
    group_type = GroupType['detector']

    def __init__(self, module: DAQ_Viewer):
        self._datatoexport_saver: DataToExportSaver = None

        self._module: 'DAQ_Viewer' = module
        self._module_group: GROUP = None
        self._h5saver = None

    def update_after_h5changed(self, ):
        self._datatoexport_saver = DataToExportSaver(self.h5saver)

    def _add_module(self, where: Union[Node, str] = None, metadata={}) -> Node:
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        metadata: dict

        Returns
        -------

        """
        if where is None:
            where = self._h5saver.raw_group

        settings_xml = ET.Element('All_settings', type='group')
        settings_xml.append(ioxml.walk_parameters_to_xml(param=self._module.settings))
        if self.main_module:
            saver_xml = ET.SubElement(settings_xml, 'H5Saver', type='group')
            saver_xml.append(ioxml.walk_parameters_to_xml(param=self._h5saver.settings))

        if self._module.ui is not None:
            for ind, viewer in enumerate(self._module.viewers):
                if hasattr(viewer, 'roi_manager'):
                    roi_xml = ET.SubElement(settings_xml, f'ROI_Viewer_{ind:02d}', type='group')
                    roi_xml.append(ioxml.walk_parameters_to_xml(param=viewer.roi_manager.settings))

        return self._h5saver.add_det_group(where, title=self._module.title, settings_as_xml=ET.tostring(settings_xml),
                                           metadata=metadata)

    def add_data(self, where: Union[Node, str], data: DataToExport):
        self._datatoexport_saver.add_data(where, data)

    def add_bkg(self, where: Union[Node, str], data_bkg: DataToExport):
        """ Adds a DataToExport as a background node in the h5file

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data_bkg: DataToExport
            The data to be saved as background

        Returns
        -------

        """
        self._datatoexport_saver.add_bkg(where, data_bkg)

    def add_external_h5(self, other_h5data: H5SaverLowLevel):
        if other_h5data is not None:
            external_group = self._h5saver.add_group('external_data', 'external_h5', self.module_group)
            try:
                if not other_h5data.isopen:
                    h5saver = H5SaverLowLevel()
                    h5saver.init_file(addhoc_file_path=other_h5data.filename)
                    h5_file = h5saver.h5_file
                else:
                    h5_file = other_h5data
                h5_file.copy_children(h5_file.get_node('/'), external_group, recursive=True)
                h5_file.flush()
                h5_file.close()

            except Exception as e:
                self.logger.exception(str(e))


class DetectorEnlargeableSaver(DetectorSaver):
    """Implementation of the ModuleSaver class dedicated to DAQ_Viewer modules in order to save enlargeable data

    Parameters
    ----------
    module
    """
    group_type = GroupType['detector']

    def __init__(self, module: DAQ_Viewer):
        super().__init__(module)
        self._datatoexport_saver: DataToExportTimedSaver = None

    def update_after_h5changed(self, ):
        self._datatoexport_saver = DataToExportTimedSaver(self.h5saver)


class DetectorExtendedSaver(DetectorSaver):
    """Implementation of the ModuleSaver class dedicated to DAQ_Viewer modules in order to save enlargeable data

    Parameters
    ----------
    module
    """
    group_type = GroupType['detector']

    def __init__(self, module: DAQ_Viewer, extended_shape: Tuple[int]):
        super().__init__(module)
        self._extended_shape = extended_shape
        self._datatoexport_saver: DataToExportExtendedSaver = None

    def update_after_h5changed(self, ):
        self._datatoexport_saver = DataToExportExtendedSaver(self.h5saver, self._extended_shape)

    def add_data(self, where: Union[Node, str], data: DataToExport, indexes: Tuple[int],
                 distribution=DataDistribution['uniform']):
        self._datatoexport_saver.add_data(where, data, indexes=indexes, distribution=distribution)

    def add_nav_axes(self, where: Union[Node, str], axes: List[Axis]):
        self._datatoexport_saver.add_nav_axes(where, axes)


class ActuatorSaver(ModuleSaver):
    """Implementation of the ModuleSaver class dedicated to DAQ_Move modules

    Parameters
    ----------
    h5saver
    module
    """
    group_type = GroupType['actuator']

    def __init__(self, module: DAQ_Move):
        self._datatoexport_saver: DataToExportTimedSaver = None
        self._module_group: GROUP = None
        self._module: DAQ_Move = module
        self._h5saver = None

    def update_after_h5changed(self, ):
        self._datatoexport_saver = DataToExportTimedSaver(self.h5saver)

    def _add_module(self, where: Union[Node, str] = None, metadata={}):
        if where is None:
            where = self._h5saver.raw_group

        settings_xml = ET.Element('All_settings')
        settings_xml.append(ioxml.walk_parameters_to_xml(param=self._module.settings))

        return self._h5saver.add_act_group(where, title=self._module.title, settings_as_xml=ET.tostring(settings_xml),
                                           metadata=metadata)

    def add_data(self, where: Union[Node, str], data: DataToExport):
        self._datatoexport_saver.add_data(where, data)


class ScanSaver(ModuleSaver):
    """Implementation of the ModuleSaver class dedicated to DAQScan module

    Parameters
    ----------
    h5saver
    module
    """
    group_type = GroupType['scan']

    def __init__(self, module):
        self._module_group: GROUP = None
        self._module: DAQScan = module
        self._h5saver = None

    def update_after_h5changed(self):
        for module in self._module.modules_manager.modules_all:
            if hasattr(module, 'module_and_data_saver'):
                module.module_and_data_saver.h5saver = self.h5saver

    def get_set_node(self, where: Union[Node, str] = None, new=False) -> GROUP:
        """Get the last group scan node

        Get the last Scan Group or create one
        get the last Scan Group if:
        * there is one already created
        * new is False

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        new: bool

        Returns
        -------
        GROUP: the GROUP associated with this module
        """
        self._module_group = self.get_last_node(where)
        new = new or (self._module_group is None)
        if new:
            self._module_group = self._add_module(where)
        for module in self._module.modules_manager.modules:
            module.module_and_data_saver.main_module = False
            module.module_and_data_saver.get_set_node(self._module_group)
        return self._module_group

    def _add_module(self, where: Union[Node, str] = None, metadata={}) -> Node:
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        metadata: dict

        Returns
        -------

        """
        if where is None:
            where = self._h5saver.raw_group

        settings_xml = ET.Element('All_settings', type='group')
        settings_xml.append(ioxml.walk_parameters_to_xml(param=self._module.settings))
        if self.main_module:
            saver_xml = ET.SubElement(settings_xml, 'H5Saver', type='group')
            saver_xml.append(ioxml.walk_parameters_to_xml(param=self._h5saver.settings))

        return self._h5saver.add_scan_group(where, title=self._module.title,
                                            settings_as_xml=ET.tostring(settings_xml),
                                            metadata=metadata)

    def add_nav_axes(self, axes: List[Axis]):
        for detector in self._module.modules_manager.detectors:
            detector.module_and_data_saver.add_nav_axes(self._module_group, axes)

    def add_data(self, dte: DataToExport = None, indexes: Tuple[int] = None,
                 distribution=DataDistribution['uniform']):
        for detector in self._module.modules_manager.detectors:
            try:
                detector.insert_data(indexes, where=self._module_group, distribution=distribution)
            except Exception as e:
                pass


class LoggerSaver(ScanSaver):
    """Implementation of the ModuleSaver class dedicated to H5Logger module

    H5Logger is the special logger to h5file of the DAQ_Logger extension

    Parameters
    ----------
    h5saver
    module
    """
    group_type = GroupType['data_logger']

    def add_data(self, dte: DataToExport):
        """Add data to it's corresponding control module

        The name of the control module is the DataToExport name attribute
        """
        if dte.name in self._module.modules_manager.detectors_name:
            control_module = self._module.modules_manager.detectors[
                self._module.modules_manager.detectors_name.index(dte.name)]
        elif dte.name in self._module.modules_manager.actuators_name:
            control_module = self._module.modules_manager.actuators[
                self._module.modules_manager.actuators_name.index(dte.name)]
        else:
            return

        control_module.append_data(dte=dte, where=self._module_group)



# -*- coding: utf-8 -*-
"""
Created the 05/12/2022

@author: Sebastien Weber
"""
from importlib import import_module
from pathlib import Path


from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.config import Config
from pymodaq.utils.daq_utils import get_entrypoints
from pymodaq.utils.scanner.scan_factory import ScannerFactory
from pymodaq.utils.enums import BaseEnum

logger = set_logger(get_module_name(__file__))
config = Config()


def register_scanner(parent_module_name: str = 'pymodaq.utils.scanner'):
    scanners = []
    try:
        scanner_module = import_module(f'{parent_module_name}.scanners')

        scanner_path = Path(scanner_module.__path__[0])

        for file in scanner_path.iterdir():
            if file.is_file() and 'py' in file.suffix and file.stem != '__init__':
                try:
                    scanners.append(import_module(f'.{file.stem}', scanner_module.__name__))
                except (ModuleNotFoundError, Exception) as e:
                    pass
    except ModuleNotFoundError:
        pass
    finally:
        return scanners


def register_scanners() -> list:
    scanners = register_scanner('pymodaq.utils.scanner')
    discovered_scanners_plugins = get_entrypoints('pymodaq.scanners')
    for entry in discovered_scanners_plugins:
        scanners.extend(register_scanner(entry.value))
    return scanners


register_scanners()
scanner_factory = ScannerFactory()
ScanType = BaseEnum('ScanType', ['NoScan'] + scanner_factory.scan_types())


class ScannerException(Exception):
    """Raised when there is an error related to the Scanner class (see pymodaq.da_utils.scanner)"""
    pass


class ScanInfo:
    """Container class for a given scan details

    It includes the number of steps and all the positions for the selected actuators. It also contains these positions
    as scan axes for easier use.

    Parameters
    ----------

    Nsteps: int
        Number of steps of the scan
    positions: ndarray
        multidimensional array. the first dimension has a length of Nsteps and each element is an actuator position
    positions_indexes: ndarray
        multidimensional array of Nsteps 0th dimension length where each element is the index
        of the corresponding positions within the axis_unique
    axes_unique: list of ndarray
        list of sorted (and with unique values) 1D arrays of unique positions of each defined axes
    selected_actuators: List[str]
        The actuators to be used for this scan
    kwargs: dict of other named parameters to be saved as attributes

    Attributes
    ----------
    Nsteps: int
        Number of steps of the scan
    positions: ndarray
        multidimensional array. the first dimension has a length of Nsteps and each element is an actuator position
    positions_indexes: ndarray
        multidimensional array of Nsteps 0th dimension length where each element is the index
        of the corresponding positions within the axis_unique
    axes_unique: list of ndarray
        list of sorted (and with unique values) 1D arrays of unique positions of each defined axes
    kwargs: dict of other named attributes
    """
    def __init__(self, Nsteps=0, positions=None, axes_indexes=None, axes_unique=None, selected_actuators=[],
                 **kwargs):
        self.Nsteps = Nsteps
        self.positions = positions
        self.axes_indexes = axes_indexes
        self.axes_unique = axes_unique
        self.selected_actuators = selected_actuators
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __repr__(self):
        return f'Scan of {self.selected_actuators} with {self.Nsteps} positions'







from pymodaq.extensions import get_models
import pymodaq.utils.config
from pymodaq.utils.logger import set_logger, get_module_name, get_module_name
from pymodaq.utils.gui_utils.file_io import select_file
from qtpy import QtWidgets
import sys
import os
from pymodaq.utils.parameter import ParameterTree, Parameter
from pymodaq.utils.parameter import ioxml
from pymodaq.utils import daq_utils as utils
from pathlib import Path
import pymodaq.utils.managers.preset_manager_utils  # to register move and det types

logger = set_logger(get_module_name(__file__))

# check if preset_mode directory exists on the drive
pid_path = pymodaq.utils.config.get_set_pid_path()
preset_path = pymodaq.utils.config.get_set_preset_path()
overshoot_path = pymodaq.utils.config.get_set_overshoot_path()
layout_path = pymodaq.utils.config.get_set_layout_path()

pid_models = [mod['name'] for mod in get_models()]


class PresetManager:
    def __init__(self, msgbox=False, path=None, extra_params=[], param_options=[]):

        if path is None:
            path = preset_path
        else:
            assert isinstance(path, Path)

        self.extra_params = extra_params
        self.param_options = param_options
        self.preset_path = path
        self.preset_params = None
        self.pid_type = False
        if msgbox:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText("Preset Manager?")
            msgBox.setInformativeText("What do you want to do?")
            cancel_button = msgBox.addButton(QtWidgets.QMessageBox.Cancel)
            new_button = msgBox.addButton("New", QtWidgets.QMessageBox.ActionRole)
            modify_button = msgBox.addButton('Modify', QtWidgets.QMessageBox.AcceptRole)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Cancel)
            ret = msgBox.exec()

            if msgBox.clickedButton() == new_button:
                self.set_new_preset()

            elif msgBox.clickedButton() == modify_button:
                path = select_file(start_path=self.preset_path, save=False, ext='xml')
                if path != '':
                    self.set_file_preset(str(path))
            else:  # cancel
                pass

    def set_file_preset(self, filename, show=True):
        """

        """
        status = False
        self.pid_type = False
        children = ioxml.XML_file_to_parameter(filename)
        self.preset_params = Parameter.create(title='Preset', name='Preset', type='group', children=children)
        if show:
            status = self.show_preset()
        return status

    def get_set_pid_model_params(self, model_file):
        self.preset_params.child('model_settings').clearChildren()
        model = get_models(model_file)
        if model is not None:
            params = model['class'].params
            self.preset_params.child('model_settings').addChildren(params)

    def set_new_preset(self):
        self.pid_type = False
        param = [
            {'title': 'Filename:', 'name': 'filename', 'type': 'str', 'value': 'preset_default'},
            {'title': 'Use PID as actuator:', 'name': 'use_pid', 'type': 'bool', 'value': False},
            # {'title': 'Saving options:', 'name': 'saving_options', 'type': 'group', 'children': H5Saver.params},
            {'title': 'PID models:', 'name': 'pid_models', 'type': 'list', 'visible': False,
             'limits': pid_models},
            {'title': 'Model Settings:', 'name': 'model_settings', 'type': 'group', 'visible': False, 'children': []},
        ]
        params_move = [
            {'title': 'Moves:', 'name': 'Moves', 'type': 'groupmove'}]  # PresetScalableGroupMove(name="Moves")]
        params_det = [{'title': 'Detectors:', 'name': 'Detectors',
                       'type': 'groupdet'}]  # [PresetScalableGroupDet(name="Detectors")]
        self.preset_params = Parameter.create(title='Preset', name='Preset', type='group',
                                              children=param + self.extra_params + params_move + params_det)
        try:
            for option in self.param_options:
                if 'path' in option and 'options_dict' in option:
                    self.preset_params.child(option['path']).setOpts(**option['options_dict'])
        except Exception as e:
            logger.exception(str(e))

        if len(pid_models) != 0:
            self.get_set_pid_model_params(pid_models[0])

        self.preset_params.sigTreeStateChanged.connect(self.parameter_tree_changed)

        status = self.show_preset()
        return status

    def parameter_tree_changed(self, param, changes):
        """
            Check for changes in the given (parameter,change,information) tuple list.
            In case of value changed, update the DAQscan_settings tree consequently.

            =============== ============================================ ==============================
            **Parameters**    **Type**                                     **Description**
            *param*           instance of pyqtgraph parameter              the parameter to be checked
            *changes*         (parameter,change,information) tuple list    the current changes state
            =============== ============================================ ==============================
        """
        for param, change, data in changes:
            path = self.preset_params.childPath(param)
            if change == 'childAdded':
                data[0].child('params', 'main_settings', 'module_name').setValue(data[0].child('name').value())

            elif change == 'value':

                if param.name() == 'use_pid':
                    self.preset_params.child('pid_models').show(param.value())
                    self.preset_params.child('model_settings').show(param.value())
                if param.name() == 'pid_models' and param.value() != '':
                    self.get_set_pid_model_params(param.value())
                if param.name() == 'name':
                    param.parent().child('params', 'main_settings', 'module_name').setValue(param.value())

            elif change == 'parent':
                pass

    def show_preset(self):
        """

        """
        dialog = QtWidgets.QDialog()
        vlayout = QtWidgets.QVBoxLayout()
        tree = ParameterTree()
        tree.setMinimumWidth(400)
        tree.setMinimumHeight(500)
        tree.setParameters(self.preset_params, showTop=False)

        vlayout.addWidget(tree)
        dialog.setLayout(vlayout)
        buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)

        buttonBox.addButton('Save', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        vlayout.addWidget(buttonBox)
        dialog.setWindowTitle('Fill in information about this manager')
        res = dialog.exec()

        if self.pid_type:
            path = pid_path
        else:
            path = self.preset_path

        if res == dialog.Accepted:
            # save managers parameters in a xml file
            # start = os.path.split(os.path.split(os.path.realpath(__file__))[0])[0]
            # start = os.path.join("..",'daq_scan')
            ioxml.parameter_to_xml_file(
                self.preset_params, os.path.join(path, self.preset_params.child('filename').value()))

            if not self.pid_type:
                # check if overshoot configuration and layout configuration with same name exists => delete them if yes
                file = os.path.splitext(self.preset_params.child('filename').value())[0]
                file = os.path.join(overshoot_path, file + '.xml')
                if os.path.isfile(file):
                    os.remove(file)

                file = os.path.splitext(self.preset_params.child('filename').value())[0]
                file = os.path.join(layout_path, file + '.dock')
                if os.path.isfile(file):
                    os.remove(file)

        return res == dialog.Accepted


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    # prog = PresetManager(True)
    prog = PresetManager(True)

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created the 04/11/2022

@author: Sebastien Weber
"""
import numpy as np
from numbers import Number
from typing import List, Tuple
from abc import ABCMeta, abstractmethod, abstractproperty

from pymodaq.utils.factory import ObjectFactory
from pymodaq.utils import math_utils as mutils
from pymodaq.utils.data import DataWithAxes, Axis, DataRaw, DataBase, DataDim, DataCalculated


config_processors = {
}


class DataProcessorBase(metaclass=ABCMeta):
    """Apply processing functions to signal data. This function should return a DataWithAxes.

    Attributes
    ----------
    apply_to: DataDim
        Specify on which type of data dimensionality this processor can be applied to, if only 1D:
        apply_to = DataDim['Data1D']

    """

    apply_to: DataDim = abstractproperty

    def process(self, data: DataWithAxes) -> DataWithAxes:
        return self.operate(data)

    @abstractmethod
    def operate(self, sub_data: DataWithAxes):
        pass

    @staticmethod
    def flatten_signal_dim(sub_data: DataWithAxes) -> Tuple[Tuple, np.ndarray]:
        """flattens data's ndarrays along the signal dimensions"""
        data_arrays = []
        new_shape = [sub_data.shape[ind] for ind in sub_data.nav_indexes]
        new_shape.append(np.prod([sub_data.shape[ind] for ind in sub_data.sig_indexes]))

        # for each data in subdata, apply the function, here argmax, along the flattened dimension. Then unravel the
        # possible multiple indexes (1 for 1D, 2 for 2D)
        for ind, data in enumerate(sub_data):
            data_arrays.append(data.reshape(new_shape))
        return new_shape, data_arrays

    def __call__(self, **kwargs):
        return self(**kwargs)


class DataProcessorFactory(ObjectFactory):
    def get(self, processor_name, **kwargs) -> DataProcessorBase:
        return self.create(processor_name, **kwargs)

    @property
    def functions(self):
        """Get the list of processor functions"""
        return self.keys_function(do_sort=False)

    def functions_filtered(self, dim: DataDim):
        """Get the list of processor functions that could be applied to data having a given dimensionality"""
        return [key for key in self.functions if self.get(key).apply_to >= dim]


@DataProcessorFactory.register('mean')
class MeanProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        data_arrays = [np.atleast_1d(np.mean(data, axis=sub_data.sig_indexes)) for data in sub_data]
        return sub_data.deepcopy_with_new_data(data_arrays, sub_data.sig_indexes)


@DataProcessorFactory.register('std')
class StdProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        data_arrays = [np.atleast_1d(np.std(data, axis=sub_data.sig_indexes)) for data in sub_data]
        return sub_data.deepcopy_with_new_data(data_arrays, sub_data.sig_indexes)


@DataProcessorFactory.register('sum')
class SumProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        data_arrays = [np.atleast_1d(np.sum(data, axis=sub_data.sig_indexes)) for data in sub_data]
        return sub_data.deepcopy_with_new_data(data_arrays, sub_data.sig_indexes)


@DataProcessorFactory.register('max')
class MaxProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        data_arrays = [np.atleast_1d(np.max(data, axis=sub_data.sig_indexes)) for data in sub_data]
        return sub_data.deepcopy_with_new_data(data_arrays, sub_data.sig_indexes)


@DataProcessorFactory.register('min')
class MinProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        data_arrays = [np.atleast_1d(np.min(data, axis=sub_data.sig_indexes)) for data in sub_data]
        return sub_data.deepcopy_with_new_data(data_arrays, sub_data.sig_indexes)


@DataProcessorFactory.register('argmax')
class ArgMaxProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        """Extract info from sub-DataWithAxes

        Retrieve the signal axis values of the maximum position of the data

        Notes
        -----
        For more complex processors, such as the argmin, argmax ... , one cannot use directly the numpy function
        (compared to min, max, mean...). Indeed one has to first flatten the data arrays on the signal axes, then apply
        the function on the flatten dimension, here get the indexes of the minimum along the flattened dimension (as
        a function of the eventual navigations dimensions). From this index, on then obtain as many indexes as signal
        dimensions (1 for 1D Signals, 2 for 2D signals). And we do this for as many data there is in sub_data.
        """
        new_data_arrays = []
        new_shape, flattened_arrays = self.flatten_signal_dim(sub_data)

        for dat in flattened_arrays:
            indexes = np.unravel_index(np.nanargmax(dat, len(new_shape)-1), sub_data.shape)[len(sub_data.nav_indexes):]
            # from the unraveled index, retrieve the corresponding axis value
            for ind in range(len(indexes)):
                axis_data = sub_data.get_axis_from_index(sub_data.sig_indexes[ind])[0].get_data()
                new_data_arrays.append(np.atleast_1d(axis_data[indexes[ind]]))
        return DataCalculated('processed_data', data=new_data_arrays, nav_indexes=sub_data.nav_indexes,
                              axes=[axis for axis in sub_data.axes if axis.index in sub_data.nav_indexes],
                              distribution=sub_data.distribution)


@DataProcessorFactory.register('argmin')
class ArgMinProcessor(DataProcessorBase):
    apply_to = DataDim['DataND']

    def operate(self, sub_data: DataWithAxes):
        """Extract info from sub-DataWithAxes

        Retrieve the signal axis values of the minimum position of the data

        Notes
        -----
        For more complex processors, such as the argmin, argmax ... , one cannot use directly the numpy function
        (compared to min, max, mean...). Indeed one has to first flatten the data arrays on the signal axes, then apply
        the function on the flatten dimension, here get the indexes of the minimum along the flattened dimension (as
        a function of the eventual navigations dimensions). From this index, on then obtain as many indexes as signal
        dimensions (1 for 1D Signals, 2 for 2D signals). And we do this for as many data there is in sub_data.
        """
        new_data_arrays = []
        new_shape, flattened_arrays = self.flatten_signal_dim(sub_data)

        for dat in flattened_arrays:
            indexes = np.unravel_index(np.nanargmin(dat, len(new_shape)-1), sub_data.shape)[len(sub_data.nav_indexes):]
            # from the unraveled index, retrieve the corresponding axis value
            for ind in range(len(indexes)):
                axis_data = sub_data.get_axis_from_index(sub_data.sig_indexes[ind])[0].get_data()
                new_data_arrays.append(np.atleast_1d(axis_data[indexes[ind]]))
        return DataCalculated('processed_data', data=new_data_arrays, nav_indexes=sub_data.nav_indexes,
                              axes=[axis for axis in sub_data.axes if axis.index in sub_data.nav_indexes],
                              distribution=sub_data.distribution)


@DataProcessorFactory.register('argmean')
class ArgMeanProcessor(DataProcessorBase):
    apply_to = DataDim['Data1D']

    def operate(self, sub_data: DataWithAxes):
        """Extract info from sub-DataWithAxes

        Retrieve the signal mean axis values

        Notes
        -----
        For more complex processors, such as the argmin, argmax ... , one cannot use directly the numpy function
        (compared to min, max, mean...). Indeed one has to first flatten the data arrays on the signal axes, then apply
        the function on the flatten dimension, here get the indexes of the minimum along the flattened dimension (as
        a function of the eventual navigations dimensions). From this index, on then obtain as many indexes as signal
        dimensions (1 for 1D Signals, 2 for 2D signals). And we do this for as many data there is in sub_data.
        """
        new_data_arrays = []
        new_shape, flattened_arrays = self.flatten_signal_dim(sub_data)
        values = sub_data.get_axis_from_index(sub_data.sig_indexes[0])[0].get_data()
        for dat in flattened_arrays:
            weights = dat
            new_data_arrays.append(np.atleast_1d(np.average(values, axis=len(new_shape) - 1, weights=weights)))
        return DataCalculated('processed_data', data=new_data_arrays, nav_indexes=sub_data.nav_indexes,
                              axes=[axis for axis in sub_data.axes if axis.index in sub_data.nav_indexes])


@DataProcessorFactory.register('argstd')
class ArgStdProcessor(DataProcessorBase):
    apply_to = DataDim['Data1D']

    def operate(self, sub_data: DataWithAxes):
        """Extract info from sub-DataWithAxes

        Retrieve the signal mean axis values

        Notes
        -----
        For more complex processors, such as the argmin, argmax ... , one cannot use directly the numpy function
        (compared to min, max, mean...). Indeed one has to first flatten the data arrays on the signal axes, then apply
        the function on the flatten dimension, here get the indexes of the minimum along the flattened dimension (as
        a function of the eventual navigations dimensions). From this index, on then obtain as many indexes as signal
        dimensions (1 for 1D Signals, 2 for 2D signals). And we do this for as many data there is in sub_data.
        """
        new_data_arrays = []
        new_shape, flattened_arrays = self.flatten_signal_dim(sub_data)
        values = sub_data.get_axis_from_index(sub_data.sig_indexes[0])[0].get_data()
        for dat in flattened_arrays:
            weights = dat
            w_avg = np.atleast_1d(np.average(values, axis=len(new_shape) - 1, weights=weights))
            new_data_arrays.append(np.atleast_1d(np.sqrt(
                np.sum(weights * (values - w_avg) ** 2, axis=len(new_shape) - 1) /
                np.sum(weights, axis=len(new_shape) - 1))))
        return DataCalculated('processed_data', data=new_data_arrays, nav_indexes=sub_data.nav_indexes,
                              axes=[axis for axis in sub_data.axes if axis.index in sub_data.nav_indexes])


if __name__ == '__main__':
    import copy
    processors = DataProcessorFactory()
    print('Builders:\n'
          f'{processors.builders}')

    print('Math functions:\n'
          f'{processors.functions}')

    # test 2D signals
    Nsigx = 200
    Nsigy = 100
    Nnav = 10
    x = np.linspace(-Nsigx / 2, Nsigx / 2 - 1, Nsigx)
    y = np.linspace(-Nsigy / 2, Nsigy / 2 - 1, Nsigy)

    dat = np.zeros((Nnav, Nsigy, Nsigx))
    for ind in range(Nnav):
        dat[ind] = ind * mutils.gauss2D(x, 10 * (ind - Nnav / 2), 25 / np.sqrt(2),
                                        y, 2 * (ind - Nnav / 2), 10 / np.sqrt(2))

    data = DataRaw('mydata', data=[dat], nav_indexes=(0,),
                   axes=[Axis('nav', data=np.linspace(0, Nnav-1, Nnav), index=0),
                         Axis('sigy', data=y, index=1),
                         Axis('sigx', data=x, index=2)])
    new_data = processors.get('sum', **config_processors).operate(data.isig[25:75, 75:125])
    print(new_data)
    print(new_data.data)



# -*- coding: utf-8 -*-
"""
Created the 25/10/2022

@author: Sebastien Weber
"""
from qtpy import QtWidgets
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager
from pymodaq.utils import config as configmod
from pymodaq.utils.daq_utils import get_version


config = configmod.Config()
BANNER = f'PyMoDAQ v{get_version()}\n' \
         'Main objects available for interaction:\n'\
         '* dashboard: Main Dashboard Object\n'\
         '* mods: ModuleManager of the dashboard\'s Control Modules\n'\
         '* np: numpy module\n\n'\
         'For inline plotting use: %matplotlib\n\n'


class QtConsole(RichJupyterWidget):
    """Live IPython console widget.
    .. image:: img/IPythonWidget.png
    :param custom_banner: Custom welcome message to be printed at the top of
       the console.
    """

    def __init__(self, parent=None, custom_banner=None, *args, **kwargs):
        if parent is not None:
            kwargs["parent"] = parent
        super().__init__(*args, **kwargs)
        if custom_banner is not None:
            self.banner = custom_banner
        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()

        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
        self.exit_requested.connect(stop)

    def push_variables(self, variable_dict):
        """ Given a dictionary containing name / value pairs, push those
        variables to the IPython console widget.
        :param variable_dict: Dictionary of variables to be pushed to the
            console's interactive namespace (```{variable_name: object, …}```)
        """
        self.kernel_manager.kernel.shell.push(variable_dict)


def main(init_qt=True):
    import sys

    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)
        if config('style', 'darkstyle'):
            import qdarkstyle
            app.setStyleSheet(qdarkstyle.load_stylesheet(qdarkstyle.DarkPalette))

    prog = QtConsole(style_sheet=config('style', 'syntax_highlighting'),
                     syntax_style=config('style', 'syntax_highlighting'),
                     custom_banner=BANNER,
                     )
    prog.show()
    if init_qt:
        sys.exit(app.exec_())
    return prog


if __name__ == '__main__':
    main()



from .dock import Dock, DockArea
from .list_picker import ListPicker
from .custom_app import CustomApp
from .file_io import select_file
from .widgets.table import TableView, TableModel, SpinBoxDelegate, BooleanDelegate
from .widgets.push import PushButtonIcon, EditPush, EditPushRel, EditPushInfo
from .widgets.qled import QLED
from .widgets.spinbox import QSpinBox_ro

from .utils import get_splash_sc



# -*- coding: utf-8 -*-
"""
Created the 22/01/2023

@author: Sebastien Weber
"""
import os
import sys
from typing import List, Union, Callable, Iterable

from qtpy import QtWidgets, QtCore

from pymodaq.utils.data import DataToExport, DataFromPlugins, DataDim, enum_checker
from pymodaq.utils.h5modules.data_saving import DataLoader
from pymodaq.utils.h5modules.saving import H5Saver
from pymodaq.utils.plotting.data_viewers.viewer import ViewerBase, ViewersEnum, ViewerDispatcher
from pymodaq.utils.gui_utils import Dock, DockArea


class LoaderPlotter:

    grouped_data0D_fullname = 'Grouped/Data0D'

    def __init__(self, dockarea):
        self.dockarea = dockarea
        self.dispatcher = ViewerDispatcher(dockarea, title='ViewerDispatcher')
        self._viewers: dict[str, ViewerBase] = None
        self._viewer_docks: dict[str, ViewerBase] = None
        self._viewer_types: List[ViewersEnum] = None
        self._h5saver: H5Saver = None
        self._data: DataToExport = None
        self.dataloader: DataLoader = None

    @property
    def viewers(self) -> List[ViewerBase]:
        return self.dispatcher.viewers

    def connect_double_clicked(self, slot: Callable):
        for viewer in self.viewers:
            viewer.sig_double_clicked.connect(slot)

    def disconnect(self, slot: Callable):
        for viewer in self.viewers:
            viewer.sig_double_clicked.disconnect(slot)

    def clear_viewers(self):
        self.dispatcher.remove_viewers(0)

    @property
    def h5saver(self):
        return self._h5saver

    @h5saver.setter
    def h5saver(self, h5saver: H5Saver):
        self._h5saver = h5saver
        self.dataloader = DataLoader(h5saver)

    @property
    def data(self) -> DataToExport:
        return self._data

    def load_data(self, filter_dims: List[Union[DataDim, str]] = None,
                  filter_full_names: List[str] = None, remove_navigation: bool = True,
                  group_0D=False, average_axis: int=None, average_index: int = 0,
                  last_step=False):
        """Load Data from the h5 node of the dataloader and apply some filtering/manipulation before
        plotting

        Parameters
        ----------
        filter_dims: List[Union[DataDim, str]]
            load only data with correct dims
        filter_full_names: List[str]
            load only data matching these names
        remove_navigation: bool
            if True, make navigation axes as signal axes (means DataND could be plotted on Viewer1D
            or Viewer2D by concatenation)
        group_0D: bool
            if True, group all (initial) Data0D into one DataFromPlugins
        average_axis: int or None
            which axis in the data shapes should be interpereted as the average (in general it is 0
            or None)
        average_index: int
            which step in the averaging process are we in.
        last_step: bool
            tells if this is the very last step of the (averaged) scan

        Returns
        -------
        DataToExport
        """

        self._data = DataToExport('All')
        self.dataloader.load_all('/', self._data)

        if average_axis is not None:
            self.average_axis(average_axis, average_index, last_step=last_step)

        if filter_dims is not None:
            filter_dims[:] = [enum_checker(DataDim, dim) for dim in filter_dims]
            self._data.data[:] = [data for data in self._data if data.dim in filter_dims]

        if filter_full_names is not None:
            self._data.data[:] = [data for data in self._data if data.get_full_name() in
                                  filter_full_names]

        if group_0D:  # 0D initial data
            self.group_0D_data()

        if remove_navigation:
            self.remove_navigation_axes()

        return self._data

    def average_axis(self, average_axis, average_index, last_step=False) -> None:
        """ Average the data along their average axis

        Parameters
        ----------
        average_axis: int or None
            which axis in the data shapes should be interpreted as the average
            (in general it is 0 or None)
        average_index: int
            which step in the averaging process are we in.
        last_step: bool
            tells if this is the very last step of the (averaged) scan
        """
        for ind, data in enumerate(self._data):
            current_data = data.inav[average_index, ...]
            if average_index > 0:
                if last_step:
                    data_to_append = data.inav[0:, ...].mean(axis=average_axis)
                else:
                    if average_index == 1:
                        data_to_append = data.inav[0, ...]
                    else:
                        data_to_append = data.inav[0:average_index, ...].mean(axis=average_axis)

                data_to_append.labels = [f'{label}_averaged' for label in data_to_append.labels]
                current_data.append(data_to_append)
            self._data[ind] = current_data

    def remove_navigation_axes(self):
        """Make the navigation axes as signal axes

        transforms DataND into Data1D or Data2D or error... depending the exact shape of the data
        and the number of navigation axes
        """
        for data in self._data:
            data.nav_indexes = ()
            data.transpose()  # because usual ND data should be plotted here as 2D with the nav axes as the minor
            # (horizontal)

    def group_0D_data(self):
        """Group in a single DataFromPlugins all data that are initialy Data0D

        """
        data = self._data.get_data_from_sig_axes(0)
        if len(data) > 0:
            data0D_arrays = []
            labels = []
            for dwa in data:
                data0D_arrays.extend(dwa.data)
                labels.extend([f'{dwa.get_full_name()}/{label}' for label in dwa.labels])
                self._data.remove(dwa)

            data0D = DataFromPlugins(self.grouped_data0D_fullname.split('/')[1],
                                     data=data0D_arrays, labels=labels,
                                     dim='DataND',
                                     origin=self.grouped_data0D_fullname.split('/')[0],
                                     axes=dwa.axes, nav_indexes=dwa.nav_indexes,
                                     )
            self._data.append(data0D)

    def load_plot_data(self, **kwargs):
        """Load and plot all data from the current H5Saver

        See Also
        -----
        load_data
        """

        target_at = kwargs.pop('target_at') if 'target_at' in kwargs else None
        last_step = kwargs.pop('last_step') if 'last_step' in kwargs else False

        self.load_data(**kwargs)
        self.show_data(target_at=target_at)
        if (last_step and 'average_index' in kwargs and kwargs['average_index']
                is not None):
            kwargs['last_step'] = last_step
            self.load_data(**kwargs)
            self.show_data(target_at=target_at)

    def show_data(self, **kwargs):
        """Send data to their dedicated viewers
        """
        #self._init_show_data(self._data)
        self.set_data_to_viewers(self._data, **kwargs)

    def _init_show_data(self, data: DataToExport):
        """Processing before showing data
        """
        self._viewer_types = [ViewersEnum(data.dim.name) for data in data]
        self.prepare_viewers(self._viewer_types)

    def prepare_viewers(self, viewers_enum: List[ViewersEnum], viewers_name: List[str] = None):
        if self._viewers is not None:
            while len(self._viewers) > 0:
                self._viewers.pop(list(self._viewers.keys())[0])
                self._viewer_docks.pop(list(self._viewer_docks.keys())[0])

        self._viewer_types = [enum_checker(ViewersEnum, viewer_enum) for viewer_enum in viewers_enum]
        if viewers_name is None or len(viewers_enum) != len(viewers_name):
            viewers_name = [f'DataPlot{ind:02d}' for ind in range(len(self._viewer_types))]

        if self.dispatcher.viewer_types != self._viewer_types:
            self.dispatcher.update_viewers(self._viewer_types)

        self._viewers = dict(zip(viewers_name, self.dispatcher.viewers))
        self._viewer_docks = dict(zip(viewers_name, self.dispatcher.viewer_docks))

    def set_data_to_viewers(self, data: DataToExport, temp=False,
                            target_at: Iterable[float] = None):
        """Process data dimensionality and send appropriate data to their data viewers

        Parameters
        ----------
        data: list of DataFromPlugins
        temp: bool
            if True notify the data viewers to display data as temporary (meaning not exporting processed data from roi)
        target_at: Iterable[float]
            if specified show and plot the roi_target of each viewer at the given position
        See Also
        --------
        ViewerBase, Viewer0D, Viewer1D, Viewer2D
        """
        for ind, _data in enumerate(data.data):
            viewer = self._viewers[_data.get_full_name()]
            self._viewer_docks[_data.get_full_name()].setTitle(_data.name)

            # viewer = self.viewers[ind]
            # self.dispatcher.viewer_docks[ind].setTitle(_data.name)

            viewer.title = _data.name
            if temp:
                viewer.show_data_temp(_data)
            else:
                viewer.show_data(_data)
            if target_at is not None:
                viewer.show_roi_target(True)
                if _data.dim == 'Data1D':
                    viewer.move_roi_target(target_at)
                elif _data.dim == 'Data2D' and _data.distribution == 'uniform':
                    _target_at = target_at.copy()

                    size = [_data.get_axis_from_index(1)[0].scaling]
                    if len(_target_at) == 1:  # means concatenation of 1D data
                        axis = _data.get_axis_from_index(0)[0]
                        size.append(axis.scaling * axis.size)
                        _target_at = list(_target_at) + [axis.offset]
                    else:
                        size.append(_data.get_axis_from_index(0)[0].scaling)
                    viewer.move_roi_target(_target_at, size)

def main(init_qt=True):
    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)

    path = r'C:\Users\weber\Downloads\temp_data.h5'

    h5saver = H5Saver()
    h5saver.open_file(path)

    win = QtWidgets.QMainWindow()
    area = DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Viewer')
    win.show()

    loader = LoaderPlotter(area)
    loader.h5saver = h5saver
    data = loader.load_data(filter_dims=['Data2D', 'Data1D'], group_1D=True)
    loader._init_show_data(data)
    loader.show_data()

    if init_qt:
        sys.exit(app.exec_())
    return loader, win


if __name__ == '__main__':
    main()
